from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import colors


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class BaseOverlayWidget(widgets.Widget):
    """Border which surrounds a widget and follows it."""

    def __init__(self, parent, **kwargs):
        super().__init__(parent, **kwargs)
        self.setAttribute(constants.WidgetAttribute.WA_TransparentForMouseEvents)
        self.setAttribute(constants.WidgetAttribute.WA_NoSystemBackground)
        self.setAttribute(constants.WidgetAttribute.WA_NoChildEventsForParent)
        # self.set_flags(tool=True)
        self._do_resize()
        self._border_width = 4
        self._border_color = gui.Color(255, 0, 0)
        self._fill_color = None
        parent.installEventFilter(self)

    def set_fill_color(self, color: datatypes.ColorType):
        self._fill_color = colors.get_color(color)

    def set_border_color(self, color: datatypes.ColorType):
        self._border_color = colors.get_color(color)

    def get_fill_color(self) -> gui.Color:
        return self._fill_color

    def get_border_color(self) -> gui.Color:
        return self._border_color

    # def event(self, event):
    #     if event.type() != core.QEvent.Type.LayoutRequest:
    #         return super().event(event)
    #     self._do_resize()
    #     return True

    # def showEvent(self, event):
    #     super().showEvent(event)
    #     # Force immediate re-layout on show
    #     self._do_resize()

    def eventFilter(self, source, event):
        match event.type():
            case core.QEvent.Type.Resize | core.QEvent.Type.Move:
                self._do_resize()
            # case core.QEvent.Type.ChildAdded:
            #     self._do_resize()
            # case core.QEvent.Type.Show:
            #     self.show()
            # case core.QEvent.Type.Hide:
            #     self.hide()
        return False

    def _do_resize(self):
        parent = self.parent()
        self.setGeometry(0, 0, parent.width(), parent.height())

    border_color = core.Property(
        gui.QColor,
        get_border_color,
        set_border_color,
        doc="Color to use for the border",
    )
    fill_color = core.Property(
        gui.QColor,
        get_fill_color,
        set_border_color,
        doc="Color to use for filling",
    )


class OverlayBorder(BaseOverlayWidget):
    def paintEvent(self, ev):
        with gui.Painter(self) as p:
            pen = gui.Pen(self._border_color, self._border_width)
            p.setPen(pen)
            if self._fill_color:
                p.setBrush(gui.Brush(self._fill_color))
            p.drawRect(self.geometry())


class FocusWidget(BaseOverlayWidget):
    def __init__(self, parent, focus_widget, overlay_color=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._parent = parent
        self._focus_widget = focus_widget
        self._fill_color = overlay_color or gui.Color(0, 0, 0, 150)

    def paintEvent(self, ev):
        with gui.Painter(self) as p:
            pen = gui.Pen(gui.Color(255, 255, 255, 0), 0)
            p.setPen(pen)
            p.setBrush(gui.Brush(self._fill_color))
            inside_rect = gui.Region(self._focus_widget.geometry())
            outside_rect = gui.Region(self.geometry())
            to_draw = outside_rect.subtracted(inside_rect)
            data_path = gui.PainterPath()
            data_path.set_fill_rule("winding")
            data_path.addRegion(to_draw)
            p.drawPath(data_path)


if __name__ == "__main__":
    app = widgets.app()
    container = widgets.Widget()
    container.set_layout("horizontal")
    widget = widgets.PlainTextEdit()
    widget2 = widgets.PlainTextEdit()
    container.box.add(widget)
    container.box.add(widget2)
    errorbox = FocusWidget(container, widget2)
    errorbox.show()
    container.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import multimedia
from prettyqt.utils import bidict


PositionStr = Literal["unspecified", "back", "front"]

POSITION: bidict[PositionStr, multimedia.QCameraDevice.Position] = bidict(
    unspecified=multimedia.QCameraDevice.Position.UnspecifiedPosition,
    back=multimedia.QCameraDevice.Position.BackFace,
    front=multimedia.QCameraDevice.Position.FrontFace,
)


class CameraDevice(multimedia.QCameraDevice):
    def get_position(self) -> PositionStr:
        return POSITION.inverse[self.position()]

    def get_video_formats(self) -> list[multimedia.CameraFormat]:
        return [multimedia.CameraFormat(i) for i in self.videoFormats()]


if __name__ == "__main__":
    fmt = CameraDevice()



"""Provides access to position, satellite info and area monitoring classes."""

from __future__ import annotations

from prettyqt.qt.QtPositioning import *  # noqa: F403

from .geoaddress import GeoAddress
from .geocoordinate import GeoCoordinate
from .geopositioninfo import GeoPositionInfo
from .geoshape import GeoShape, GeoShapeMixin
from .georectangle import GeoRectangle
from .geocircle import GeoCircle
from .geopath import GeoPath
from .geopolygon import GeoPolygon
from .geolocation import GeoLocation
from .geosatelliteinfo import GeoSatelliteInfo
from .geopositioninfosource import GeoPositionInfoSource, GeoPositionInfoSourceMixin
from .geosatelliteinfosource import GeoSatelliteInfoSource
from .nmeapositioninginfosource import NmeaPositionInfoSource
from .geoareamonitorinfo import GeoAreaMonitorInfo
from .geoareamonitorsource import GeoAreaMonitorSource
from prettyqt.qt import QtPositioning

QT_MODULE = QtPositioning

__all__ = [
    "GeoAddress",
    "GeoAreaMonitorInfo",
    "GeoAreaMonitorSource",
    "GeoCircle",
    "GeoCoordinate",
    "GeoLocation",
    "GeoPath",
    "GeoPolygon",
    "GeoPositionInfo",
    "GeoPositionInfoSource",
    "GeoPositionInfoSourceMixin",
    "GeoRectangle",
    "GeoSatelliteInfo",
    "GeoSatelliteInfoSource",
    "GeoShape",
    "GeoShapeMixin",
    "NmeaPositionInfoSource",
]



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

from prettyqt import core, gui


if TYPE_CHECKING:
    from collections.abc import Callable


class UndoStackMixin(core.ObjectMixin):
    def __len__(self) -> int:
        return self.count()

    def __getitem__(self, index: int) -> gui.QUndoCommand:
        cmd = self.command(index)
        if cmd is None:
            raise KeyError(index)
        return cmd

    @contextlib.contextmanager
    def create_macro(self, text: str):
        self.beginMacro(text)
        yield None
        self.endMacro()

    def add_command(self, title: str, redo: Callable, undo: Callable) -> gui.UndoCommand:
        class MyCommand(gui.UndoCommand):
            def redo(self):
                return redo()

            def undo(self):
                return undo()

        cmd = MyCommand(title)
        self.push(cmd)
        return cmd


class UndoStack(UndoStackMixin, gui.QUndoStack):
    pass



from __future__ import annotations

from prettyqt.qt import QtCore


class Property(QtCore.Property):
    """Template class that enables automatic property bindings."""

    def __init__(self, *args, **kwargs):
        self.doc = kwargs.get("doc")
        super().__init__(*args, **kwargs)

    @classmethod
    def get_doc_dict(cls, klass: type):
        import inspect

        return {
            name: member.doc
            for name, member in inspect.getmembers(klass)
            if isinstance(member, cls) and hasattr(member, "doc")
        }


if __name__ == "__main__":
    prop = Property(int)



from __future__ import annotations

from typing import TYPE_CHECKING, Self


if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence


class NestedItem:
    item_name = "not_defined"

    def __init__(
        self,
        parent: Self | None = None,
        dynamic_name: str | None = None,
        count: int | None = None,
        children: Sequence[Self] | None = None,
    ):
        self.parent = parent
        self.dynamic_name = dynamic_name or self.item_name
        self.count = count
        # self.timestamp = kwargs.pop("timestamp", time.time())
        self.children: Sequence[Self] = []
        if children:
            self.add_children(children)

    def __iter__(self) -> Iterator[Self]:
        return iter(self.children)

    def add_children(self, children: Sequence[Self]):
        for child in children:
            child.parent = self
        self.children.extend(children)

    def append_child(self, item: Self):
        item.parent_item = self
        self.children.append(item)

    def insert_children(self, idx: int, items: Sequence[Self]):
        self.children[idx:idx] = items
        for item in items:
            item.parent_item = self

    def child(self, row: int) -> Self:
        return self.children[row]

    def row(self) -> int:
        """Return row number.

        returns row position of item inside parent`s children
        returns 0 if no parent available

        Returns:
            row number
        """
        return self.parent.children.index(self) if self.parent else 0

    def iter_tree(
        self,
        name: str | None = None,
        yield_self: bool = True,
        recursive: bool = True,
        level: int = 0,
        count: int = 0,
        assign_names: bool = True,
    ):
        """Yield children from bottom to top, yield self at end.

        assigns dynamic vars count, level and dynamic name
        """
        if name is None:
            name = self.item_name
        for i, c in enumerate(self.children, start=1):
            count += 1
            level += 1
            with_suffix = f"{name}_{i}" if len(self.children) > 1 else name
            if recursive:
                yield from c.iter_tree(
                    with_suffix, yield_self=False, level=level, count=count
                )
            if assign_names:
                c.dynamic_name, c.count = with_suffix, count
            yield c
            level -= 1
        if yield_self:
            if assign_names:
                self.dynamic_name, self.count = name, 0
            yield self



"""Provides QtGui classes and functions."""

from prettyqt.qt import PYQT6, PYSIDE6
from prettyqt.utils import get_repr

if PYQT6:
    from PyQt6.QtGui import *

    for cls in (QPalette,):
        for attr in dir(cls):
            if not attr[0].isupper():
                continue
            ns = getattr(cls, attr)
            for name, val in vars(ns).items():
                if not name.startswith("_"):
                    setattr(cls, name, val)

    def pos(self, *a):
        _pos = self.position(*a)
        return _pos.toPoint()

    QMouseEvent.pos = pos  # type: ignore

    # mightBeRichText is strangely in QtGui.Qt for PySide6..
    from PyQt6.QtCore import Qt


elif PYSIDE6:
    from PySide6.QtGui import *  # type: ignore
else:
    raise ModuleNotFoundError("No Qt bindings could be found")


def __repr__(self):
    return get_repr(self, self.red(), self.green(), self.blue(), self.alpha())


QColor.__repr__ = __repr__



from __future__ import annotations

from collections.abc import Hashable
import logging
import pathlib
from typing import TYPE_CHECKING, ClassVar

from prettyqt import constants, itemmodels


if TYPE_CHECKING:
    from prettyqt import core


logger = logging.getLogger(__name__)


class TupleTreeModel(itemmodels.TreeModel):
    """Base Tree Model to display a dict[tuple[Any], str] data structure.

    The dictionary keys are tuples of path parts, like ("path", "to", "something")

    ```py
    model = TupleTreeModel({("a"): "test2", ("a", "b"): "test", ("a", "b", "c"): "test3"})
    table = widgets.TreeView()
    table.set_model(model)
    table.show()
    ```
    """

    SUPPORTS = dict[tuple[Hashable, ...] | pathlib.Path, str]
    HEADER: ClassVar = ["Name"]

    def __init__(self, mapping: dict, **kwargs):
        super().__init__((), **kwargs)
        self.mapping = {(): "root"} | mapping

    def columnCount(self, parent: core.ModelIndex | None = None):
        return len(self.HEADER)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role, section:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE, _:
                return self.HEADER[section]
        return None

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        tup = self.data_by_index(index).obj
        match role, index.column():
            case constants.DISPLAY_ROLE, 0:
                return str(self.mapping[tup])

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case dict() if all(
                isinstance(k, pathlib.Path | tuple) for k in instance.keys() and instance
            ):
                return True
        return False

    def _fetch_object_children(
        self, item: TupleTreeModel.TreeItem
    ) -> list[TupleTreeModel.TreeItem]:
        parts = item.obj.parts if isinstance(item.obj, pathlib.Path) else item.obj
        return [
            self.TreeItem(obj=k)
            for k in self.mapping
            if len(k) == len(parts) + 1
            and all(parent_part == k[i] for i, parent_part in enumerate(parts))
        ]

    def _has_children(self, item: TupleTreeModel.TreeItem) -> bool:
        parts = item.obj.parts if isinstance(item.obj, pathlib.Path) else item.obj
        return any(
            len(k) == len(parts) + 1
            and all(parent_part == k[i] for i, parent_part in enumerate(parts))
            for k in self.mapping
        )


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    view = widgets.TreeView()
    view.setRootIsDecorated(True)
    mapping = {"a": "test", ("a", "b"): "test", ("a", "b", "c"): "test"}
    model = TupleTreeModel(mapping)
    view.set_model(model)
    view.setEditTriggers(view.EditTrigger.AllEditTriggers)
    view.set_delegate("editor")
    view.resize(1000, 1000)
    view.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtScxml


class ScxmlDataModelMixin(core.ObjectMixin):
    pass


class ScxmlDataModel(ScxmlDataModelMixin, QtScxml.QScxmlDataModel):
    """The data model base class for a Qt SCXML state machine."""



from __future__ import annotations

from collections.abc import Callable
import dataclasses
import logging
import string

from prettyqt import constants, core


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class ColumnMapping:
    formatter: str | Callable
    header: str
    flags: constants.ItemFlag | None = None


class ColumnJoinerProxyModel(core.AbstractProxyModel):
    """Proxy model which joins the contents of several columns.

    The columns are joined based on a formatter and appended to the end of the model
    as a new column.

    The formatter must look like `{0} - {1}: {4}`.
    The format codes are then populated with the content of given columns,
    in this case it would be `{Text of column 0} - {Text of Column 1}: {Text of Column 4}`

    ### Example

    ```py
    table.proxifier.join_columns(formatter="{0} - {2}", header="New column")
    table.show()
    # or
    model = MyModel()
    proxy = ColumnJoinerProxyModel()
    proxy.set_source_model(model)
    proxy.add_column(formatter="{0} - {2}", header="New column")
    proxy.add_column(formatter="{4}: {5}", header="Another column")
    table.set_model(proxy)
    table.show()
    ```

    === "Without proxy"

        ```py
        data = dict(first=["John", "Mary"], last=["Doe", "Bo"])
        model = gui.StandardItemModel.from_dict(data)
        table = widgets.TableView()
        table.set_model(model)
        # table.proxifier.add_column(header="Full name", formatter="{1}, {0}")
        table.show()
        ```
        <figure markdown>
          ![Image title](../../images/columnjoinerproxymodel_before.png)
        </figure>

    === "With proxy"

        ```py
        data = dict(first=["John", "Mary"], last=["Doe", "Bo"])
        model = gui.StandardItemModel.from_dict(data)
        table = widgets.TableView()
        table.set_model(model)
        table.proxifier.add_column(header="Full name", formatter="{1}, {0}")
        table.show()
        ```
        <figure markdown>
          ![Image title](../../images/columnjoinerproxymodel_after.png)
        </figure>
    """

    ID = "column_join"
    ICON = "mdi.table-column-plus-before"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mapping = []

    def columnCount(self, parent: core.ModelIndex | None = None) -> int:
        parent = parent or core.ModelIndex()
        return (
            0
            if self.sourceModel() is None
            else self.sourceModel().columnCount(parent) + len(self.mapping)
        )

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        return self.sourceModel().rowCount()

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        column = index.column()
        if self.is_additional_column(column):
            flags = self.mapping[column - self.columnCount()].flags
            return (
                flags
                if flags is not None
                else constants.IS_ENABLED | constants.IS_SELECTABLE
            )
        return self.sourceModel().flags(index)

    def is_additional_column(self, column: int):
        col_count = self.sourceModel().columnCount()
        return column >= col_count

    def index(self, row: int, column: int, parent: core.ModelIndex | None = None):
        parent = parent or core.ModelIndex()
        if self.is_additional_column(column):
            return self.createIndex(row, column, core.ModelIndex())
        return self.sourceModel().index(row, column, parent)

    def parent(self, index=None):
        if index is None:
            return super().parent()
        if self.is_additional_column(index.column()):
            return core.ModelIndex()
        return self.sourceModel().parent(index)

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        col_count = self.sourceModel().columnCount()
        column = index.column()
        if self.is_additional_column(column):
            if role == constants.DISPLAY_ROLE:
                mapper = self.mapping[column - col_count]
                formatter = mapper.formatter
                field_names = [
                    field_name
                    for (
                        _literal_text,
                        field_name,
                        _format_spec,
                        _conversion,
                    ) in string.Formatter().parse(formatter)
                    # if field_name is not None
                ]
                match formatter:
                    case str():
                        for name in field_names:
                            idx = self.index(index.row(), int(name), index.parent())
                            data = self.data(idx)
                            formatter = formatter.replace(f"{{{name}}}", data)
                        return formatter
                    case Callable():
                        return formatter(index)
            return None
        return super().data(index, role)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if orientation == constants.HORIZONTAL and self.is_additional_column(section):
            if role == constants.DISPLAY_ROLE:
                mapper = self.mapping[section - self.columnCount()]
                return mapper.header
            return None
        return super().headerData(section, orientation, role)

    def mapToSource(self, proxy_index):
        if not proxy_index.isValid():
            return core.ModelIndex()
        column = proxy_index.column()
        if self.is_additional_column(column):
            return core.ModelIndex()
        return self.sourceModel().index(proxy_index.row(), column, proxy_index.parent())

    def mapFromSource(self, index):
        if self.is_additional_column(index.column()):
            return core.ModelIndex()
        return self.sourceModel().index(index.row(), index.column(), index.parent())

    def add_mapping(
        self, header: str, formatter: str, flags: constants.ItemFlag | None = None
    ):
        """Add a new column to the table.

        Arguments:
            header: Label used for the section header.
            formatter: String formatter (example "{0}: {1}")
            flags: ItemFlags for new column
        """
        self.mapping.append(ColumnMapping(formatter, header, flags))


if __name__ == "__main__":
    from prettyqt import gui, widgets

    app = widgets.app()
    data = dict(first=["John", "Mary"], last=["Doe", "Bo"])
    model = gui.StandardItemModel.from_dict(data)
    table = widgets.TableView()
    table.set_model(model)
    table.proxifier.add_column(header="Full name", formatter="{1}, {0}")
    table.show()
    table.resize(400, 130)
    table.h_header.resize_sections("stretch")
    table.set_title("ColumnJoinerProxymodel")
    table.set_icon("mdi.table-column-plus-before")
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

import prettyqt
from prettyqt import charts, gui
from prettyqt.utils import datatypes


if TYPE_CHECKING:
    from prettyqt import core


class XYSeriesMixin(charts.AbstractSeriesMixin):
    def __setitem__(self, index: int, val: datatypes.PointFType):
        """Set point at given index to value."""
        self.replace(index, datatypes.to_pointf(val))

    def __delitem__(self, index: int):
        """Remove point with given index."""
        self.remove(index)

    # def __setstate__(self, state):
    #     self.append(state["points"])

    # def __reduce__(self):
    #     return type(self), (), self.__getstate__()

    def __add__(self, other: datatypes.PointFType) -> XYSeries:
        """Append a point to the Series."""
        self.append(datatypes.to_pointf(other))
        return self

    def get_points(self) -> list[core.QPoint]:
        if prettyqt.qt.API == "pyqt6":
            return [self.at(i) for i in range(self.count())]
        return self.points()

    def get_pen(self) -> gui.Pen:
        return gui.Pen(self.pen())

    def get_brush(self) -> gui.Brush:
        return gui.Brush(self.brush())


class XYSeries(XYSeriesMixin, charts.QXYSeries):
    pass



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Literal

from prettyqt import core


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


logger = logging.getLogger(__name__)

NameFormatStr = Literal["hex_rgb", "hex_argb"]
NameStr = NameFormatStr | Literal["svg_rgb", "svg_argb", "qcss_rgb", "qcss_argb"]


class TextAnimation(core.PropertyAnimation):
    ID = "text"

    def __init__(
        self,
        start: datatypes.VariantType | None = None,
        end: datatypes.VariantType | None = None,
        duration: int = 1000,
        easing: core.easingcurve.TypeStr = "in_out_sine",
        fmt: NameStr = "html",
        mask: str = "{text}",
        parent: core.QObject | None = None,
    ):
        self._fmt = fmt
        self._mask = mask
        super().__init__(parent)
        self._child_anim = core.VariantAnimation()
        self._child_anim.valueChanged.connect(self.updateCurrentValue)
        self.set_easing(easing)
        self.set_duration(duration)
        if start is not None:
            self._child_anim.set_start_value(start)
        if end is not None:
            self._child_anim.set_end_value(end)
        self.setStartValue("")
        self.setStartValue("")

    def start(self, *args, **kwargs):
        self._child_anim.start(*args, **kwargs)
        super().start(*args, **kwargs)

    # TODO: also do stop etc.

    def set_start_value(self, value: datatypes.VariantType):
        self._child_anim.set_start_value(value)

    def set_end_value(self, value: datatypes.VariantType):
        self._child_anim.set_end_value(value)

    def set_easing(self, easing: core.easingcurve.TypeStr | core.QEasingCurve.Type):
        super().set_easing(easing)
        self._child_anim.set_easing(easing)

    def set_duration(self, duration: int):
        super().setDuration(duration)
        self._child_anim.set_duration(duration)

    def updateCurrentValue(self, value: datatypes.VariantType):
        match value:
            case None:
                return None
            case gui.QColor():
                color = gui.Color(value)
                text = color.get_name(self._fmt) if self._fmt else color.toString()
                value = self._mask.format(text=text)
            case int() | float():
                value = self._mask.format(text=str(value))
            case core.QPoint() | core.QPointF():
                return f"{value.x()}, {value.y()}"
            case core.QSize() | core.QSizeF():
                return f"{value.width()}, {value.width()}"
        logger.debug(value)
        super().updateCurrentValue(value)
        return None


if __name__ == "__main__":
    from prettyqt import gui, widgets

    app = widgets.app()
    w = widgets.RadioButton()
    mask = "QWidget {{background-color:{text};}}"
    anim = TextAnimation(
        gui.QColor("red"), gui.QColor("blue"), parent=w, mask=mask, fmt="qcss_rgb"
    )
    anim.apply_to(w.styleSheet)
    anim.start()
    w.show()
    with app.debug_mode():
        app.exec()



"""Module containing custom delegate classes.

PrettyQt offers an extended Item delegate with support for a large number of
different data types.

The following data types are supported:

* bool
* enum.Flag
* enum.Enum
* int
* float
* str
* range
* slice
* list of ints
* list of floats
* list of strings
* pathlib.Path
* re.Pattern
* datetime.date
* datetime.time
* datetime.datetime
* QtCore.QRegularExpression
* QtCore.QTime
* QtCore.QDate
* QtCore.QDateTime
* QtCore.QPoint
* QtCore.QPointF
* QtCore.QRect
* QtCore.QRectF
* QtCore.QRection
* QtCore.QKeyCombination
* QtCore.QLocale
* QtCore.QSize
* QtCore.QSizeF
* QtCore.QUrl
* QtGui.QFont
* QtGui.QKeySequence
* QtGui.QPalette
* QtGui.QColor
* QtGui.QBrush
* QtGui.QCursor
* QtGui.QIcon
* QtWidgets.QSizePolicy

If numpy is installed, the following types are supported, too:

* numpy.floating
* numpy.integer
* numpy.str_
* numpy.datetime64
* numpy.bool_

"""

from .buttondelegate import ButtonDelegate
from .htmlitemdelegate import HtmlItemDelegate, MarkdownItemDelegate
from .icondelegate import IconDelegate
from .nofocusdelegate import NoFocusDelegate
from .progressbardelegate import ProgressBarDelegate
from .radiodelegate import RadioDelegate
from .renderlinkdelegate import RenderLinkDelegate
from .stardelegate import StarDelegate
from .editordelegate import EditorDelegate
from .widgetdelegate import WidgetDelegate


__all__ = [
    "ButtonDelegate",
    "EditorDelegate",
    "HtmlItemDelegate",
    "IconDelegate",
    "MarkdownItemDelegate",
    "NoFocusDelegate",
    "ProgressBarDelegate",
    "RadioDelegate",
    "RenderLinkDelegate",
    "StarDelegate",
    "WidgetDelegate",
]



from __future__ import annotations

from prettyqt import core, widgets


class Dial(widgets.AbstractSliderMixin, widgets.QDial):
    """Rounded range control (like a speedometer or potentiometer)."""

    value_changed = core.Signal(int)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.valueChanged.connect(self.on_value_change)


if __name__ == "__main__":
    app = widgets.app()
    slider = Dial()
    slider.setRange(0, 100)
    slider.show()
    app.exec()



from __future__ import annotations

import logging

from prettyqt import core, itemmodels


logger = logging.getLogger(__name__)


class LinkedSelectionModel(core.ItemSelectionModel):
    """An ItemSelectionModel which link selections between item models."""

    # TODO: atm this doesnt need to inherit from ItemSelectionModel.
    # Not sure if there is any advantage in doing so...
    # Otherwise we could rename to SelectionLinker and just inherit from
    # object / core.Object
    def __init__(
        self,
        *itemviews,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._itemviews = itemviews
        self._models = [w.model() for w in itemviews]
        self._mapper = itemmodels.ProxyMapper(*self._models)
        for w in itemviews:
            w.selectionModel().currentChanged.connect(self._on_current_change)
            w.selectionModel().selectionChanged.connect(self._on_selection_change)

    def _on_current_change(self, new: core.ModelIndex, _):
        source = self.sender().model()
        source_index = self._models.index(source)
        target_indexes = list(range(len(self._models)))
        target_indexes.remove(source_index)
        for target_index in target_indexes:
            mapped = self._mapper.map_index(
                source=source_index, target=target_index, index=new
            )
            self._itemviews[target_index].setCurrentIndex(mapped)

    def _on_selection_change(self, new: core.QItemSelection, _):
        source_model = self.sender().model()
        source_index = self._models.index(source_model)
        target_indexes = list(range(len(self._models)))
        target_indexes.remove(source_index)
        for target_index in target_indexes:
            selected = self._mapper.map_selection(
                source=source_index, target=target_index, selection=new
            )
            sel_model = self._itemviews[target_index].selectionModel()
            sel_model.select(selected, sel_model.SelectionFlag.Select)


if __name__ == "__main__":
    from prettyqt import debugging, widgets

    app = widgets.app(style="Vista")
    table = debugging.example_table()
    table.proxifier.transpose()
    table.proxifier.get_proxy("table_to_list")
    w = debugging.ProxyComparerWidget(table.model())
    tables = w.find_children(widgets.TableView)
    with app.debug_mode():
        w.show()
        app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core, gui, widgets


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class BaseRectEdit(widgets.Widget):
    Typ: type
    Widget: type[widgets.QWidget]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_margin(0)
        self.spinbox_x = self.Widget(maximum=999999)
        self.spinbox_y = self.Widget(maximum=999999)
        self.spinbox_width = widgets.SpinBox(maximum=999999)
        self.spinbox_height = widgets.SpinBox(maximum=999999)
        layout = self.set_layout("horizontal", margin=0)
        layout.add(widgets.Label("x", alignment="center_right"))
        layout.add(self.spinbox_x)
        layout.add(widgets.Label("y", alignment="center_right"))
        layout.add(self.spinbox_y)
        layout.add(widgets.Label("w", alignment="center_right"))
        layout.add(self.spinbox_width)
        layout.add(widgets.Label("h", alignment="center_right"))
        layout.add(self.spinbox_height)

        self.spinbox_x.value_changed.connect(self._on_value_change)
        self.spinbox_y.value_changed.connect(self._on_value_change)
        self.spinbox_width.value_changed.connect(self._on_value_change)
        self.spinbox_height.value_changed.connect(self._on_value_change)

    def _on_value_change(self):
        self._value = self.get_value()
        self.value_changed.emit(self._value)

    def get_value(self) -> core.Rect | core.RectF:
        return self.Typ(
            self.spinbox_x.get_value(),
            self.spinbox_y.get_value(),
            self.spinbox_width.get_value(),
            self.spinbox_height.get_value(),
        )

    def set_value(self, value: datatypes.RectType | datatypes.RectFType):
        if isinstance(value, tuple):
            value = self.Typ(*value)
        self._value = value
        self.spinbox_x.set_value(value.x())
        self.spinbox_y.set_value(value.y())
        self.spinbox_width.set_value(value.width())
        self.spinbox_height.set_value(value.height())


class RectEdit(BaseRectEdit):
    Typ = core.QRect
    Widget = widgets.SpinBox
    value_changed = core.Signal(core.QRect)

    def set_value(self, value: datatypes.RectType):
        super().set_value(value)

    def get_value(self) -> core.Rect:
        return super().get_value()

    value = core.Property(
        core.QRect,
        get_value,
        set_value,
        user=True,
        doc="Current value",
    )


class RectFEdit(BaseRectEdit):
    Typ = core.QRectF
    Widget = widgets.DoubleSpinBox
    value_changed = core.Signal(core.QRectF)

    def set_value(self, value: datatypes.RectFType):
        super().set_value(value)

    def get_value(self) -> core.RectF:
        return super().get_value()

    value = core.Property(
        core.QRectF,
        get_value,
        set_value,
        user=True,
        doc="Current value",
    )


class RegionEdit(BaseRectEdit):
    Widget = widgets.SpinBox
    value_changed = core.Signal(gui.QRegion)
    Typ = gui.QRegion

    def get_value(self) -> gui.Region:
        return gui.Region(
            self.spinbox_x.get_value(),
            self.spinbox_y.get_value(),
            self.spinbox_width.get_value(),
            self.spinbox_height.get_value(),
        )

    def set_value(self, value: gui.QRegion):
        if isinstance(value, tuple):
            value = gui.Region(*value)
        self._value = value
        bounding_rect = value.boundingRect()
        self.spinbox_x.set_value(bounding_rect.x())
        self.spinbox_y.set_value(bounding_rect.y())
        self.spinbox_width.set_value(bounding_rect.width())
        self.spinbox_height.set_value(bounding_rect.height())

    value = core.Property(
        gui.Region,
        get_value,
        set_value,
        user=True,
        doc="Current value",
    )


if __name__ == "__main__":
    app = widgets.app()
    widget = RectFEdit(window_title="Test")
    widget.set_value(core.QRectF())
    widget.value_changed.connect(print)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.utils import bidict


CORNER_WIDGETS = bidict(
    none=widgets.QStyleOptionTab.CornerWidget.NoCornerWidgets,
    left=widgets.QStyleOptionTab.CornerWidget.LeftCornerWidget,
    right=widgets.QStyleOptionTab.CornerWidget.RightCornerWidget,
)

SELECTED_POSITION = bidict(
    not_adjacent=widgets.QStyleOptionTab.SelectedPosition.NotAdjacent,
    next_is_selected=widgets.QStyleOptionTab.SelectedPosition.NextIsSelected,
    previous_is_selected=widgets.QStyleOptionTab.SelectedPosition.PreviousIsSelected,
)

TAB_FEATURE = bidict(
    none=widgets.QStyleOptionTab.TabFeature(0),  # type: ignore
    has_frame=widgets.QStyleOptionTab.TabFeature.HasFrame,
)

TAB_POSITION = bidict(
    beginning=widgets.QStyleOptionTab.TabPosition.Beginning,
    middle=widgets.QStyleOptionTab.TabPosition.Middle,
    end=widgets.QStyleOptionTab.TabPosition.End,
    only_one_tab=widgets.QStyleOptionTab.TabPosition.OnlyOneTab,
)


class StyleOptionTab(widgets.StyleOptionMixin, widgets.QStyleOptionTab):
    pass



from __future__ import annotations

import os
import pathlib
from typing import TYPE_CHECKING, Literal, Self

from prettyqt import core
from prettyqt.utils import bidict, get_repr


if TYPE_CHECKING:
    from collections.abc import Iterable

    from prettyqt.utils import datatypes


FilterStr = Literal[
    "none",
    "dirs",
    "all_dirs",
    "files",
    "drives",
    "no_sym_links",
    "no_dot_and_dotdot",
    "no_dot",
    "no_dotdot",
    "all_entries",
    "readable",
    "writable",
    "executable",
    "modified",
    "hidden",
    "system",
    "case_sensitive",
]

FILTERS: bidict[FilterStr, core.QDir.Filter] = bidict(
    none=core.QDir.Filter.NoFilter,
    dirs=core.QDir.Filter.Dirs,
    all_dirs=core.QDir.Filter.AllDirs,
    files=core.QDir.Filter.Files,
    drives=core.QDir.Filter.Drives,
    no_sym_links=core.QDir.Filter.NoSymLinks,
    no_dot_and_dotdot=core.QDir.Filter.NoDotAndDotDot,
    no_dot=core.QDir.Filter.NoDot,
    no_dotdot=core.QDir.Filter.NoDotDot,
    all_entries=core.QDir.Filter.AllEntries,
    readable=core.QDir.Filter.Readable,
    writable=core.QDir.Filter.Writable,
    executable=core.QDir.Filter.Executable,
    modified=core.QDir.Filter.Modified,
    hidden=core.QDir.Filter.Hidden,
    system=core.QDir.Filter.System,
    case_sensitive=core.QDir.Filter.CaseSensitive,
)

SortFlagStr = Literal[
    "name",
    "time",
    "size",
    "type",
    "unsorted",
    "no_sort",
    "dirs_first",
    "dirs_last",
    "reversed",
    "ignore_case",
    "locale_aware",
]

SORT_FLAG: bidict[SortFlagStr, core.QDir.SortFlag] = bidict(
    name=core.QDir.SortFlag.Name,
    time=core.QDir.SortFlag.Time,
    size=core.QDir.SortFlag.Size,
    type=core.QDir.SortFlag.Type,
    unsorted=core.QDir.SortFlag.Unsorted,
    no_sort=core.QDir.SortFlag.NoSort,
    dirs_first=core.QDir.SortFlag.DirsFirst,
    dirs_last=core.QDir.SortFlag.DirsLast,
    reversed=core.QDir.SortFlag.Reversed,
    ignore_case=core.QDir.SortFlag.IgnoreCase,
    locale_aware=core.QDir.SortFlag.LocaleAware,
)


class Dir(core.QDir):
    """Access to directory structures and their contents."""

    def __getattr__(self, attr: str):
        return getattr(self.to_path(), attr)

    def __repr__(self):
        return get_repr(self, self.absolutePath())

    def __str__(self):
        return self.absolutePath()

    def __reduce__(self):
        return type(self), (self.absolutePath(),)

    def __truediv__(self, other: datatypes.PathType) -> pathlib.Path:
        return self.to_path() / os.fspath(other)

    def __fspath__(self) -> str:
        return self.absolutePath()

    def __bool__(self):
        return self.exists()

    def __abs__(self) -> str:
        return self.absolutePath()

    @property
    def _absolutePath(self) -> str:
        return self.absolutePath()

    __match_args__ = ("_absolutePath",)

    def to_path(self) -> pathlib.Path:
        return pathlib.Path(self.absolutePath())

    def set_filter(self, *filters: FilterStr):
        flags = FILTERS.merge_flags(filters)
        self.setFilter(flags)

    def get_filter(self) -> list[FilterStr]:
        return FILTERS.get_list(self.filter())

    def get_entry_info_list(
        self, sort_mode: SortFlagStr = "no_sort", filters: FilterStr = "none"
    ) -> list[core.FileInfo]:
        return [
            core.FileInfo(i)
            for i in self.entryInfoList(
                sort=SORT_FLAG[sort_mode],
                filters=self.Filter.AllEntries | FILTERS[filters],
            )
        ]

    def get_entry_list(
        self, sort_mode: SortFlagStr = "no_sort", filters: FilterStr = "none"
    ) -> list[pathlib.Path]:
        return [
            pathlib.Path(i)
            for i in self.entryList(sort=SORT_FLAG[sort_mode], filters=FILTERS[filters])
        ]

    @classmethod
    def get_current(cls) -> Self:
        return cls(cls.current())

    @classmethod
    def get_home(cls) -> Self:
        return cls(cls.home())

    @classmethod
    def get_current_path(cls) -> pathlib.Path:
        return pathlib.Path(cls.currentPath())

    @classmethod
    def get_home_path(cls) -> pathlib.Path:
        return pathlib.Path(cls.homePath())

    @classmethod
    def get_temp_path(cls) -> pathlib.Path:
        return pathlib.Path(cls.tempPath())

    @classmethod
    def get_drives(cls) -> list[core.FileInfo]:
        return [core.FileInfo(i) for i in cls.drives()]

    @classmethod
    def add_search_path(cls, prefix: str, path: datatypes.PathType):
        cls.addSearchPath(prefix, os.fspath(path))

    @classmethod
    def set_search_paths(cls, prefix: str, paths: Iterable[datatypes.PathType]):
        cls.setSearchPaths(prefix, [os.fspath(p) for p in paths])


if __name__ == "__main__":
    path = Dir.get_temp_path()



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from prettyqt import core
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from collections.abc import Callable


DirectionStr = Literal["forward", "backward"]

DIRECTION: bidict[DirectionStr, core.QTimeLine.Direction] = bidict(
    forward=core.QTimeLine.Direction.Forward,
    backward=core.QTimeLine.Direction.Backward,
)


StateStr = Literal["not_running", "paused", "running"]

STATE: bidict[StateStr, core.QTimeLine.State] = bidict(
    not_running=core.QTimeLine.State.NotRunning,
    paused=core.QTimeLine.State.Paused,
    running=core.QTimeLine.State.Running,
)


class TimeLine(core.ObjectMixin, core.QTimeLine):
    """Timeline for controlling animations."""

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"direction": DIRECTION}
        return maps

    def set_direction(self, direction: DirectionStr | core.QTimeLine.Direction):
        """Set the direction.

        Args:
            direction: direction
        """
        self.setDirection(DIRECTION.get_enum_value(direction))

    def get_direction(self) -> DirectionStr:
        """Return current direction.

        Returns:
            direction
        """
        return DIRECTION.inverse[self.direction()]

    def get_state(self) -> StateStr:
        """Return current state.

        Returns:
            state
        """
        return STATE.inverse[self.state()]

    def set_easing(
        self,
        easing_type: core.easingcurve.TypeStr
        | core.QEasingCurve.Type
        | Callable[[float], float],
    ) -> core.EasingCurve:
        curve = core.EasingCurve()
        if callable(easing_type):
            curve.set_custom_type(easing_type)
        else:
            curve.set_type(easing_type)
        self.setEasingCurve(curve)
        return curve

    def get_easing(self) -> core.easingcurve.TypeStr | Callable[[float], float]:
        curve = core.EasingCurve(self.easingCurve())
        typ = curve.get_type()
        return curve.get_custom_type() if typ == "custom" else typ


if __name__ == "__main__":
    timeline = TimeLine(finished=print, value_changed=print)



from __future__ import annotations

from prettyqt import constants, core
from prettyqt.qt import QtGui


class FontMetricsF(QtGui.QFontMetricsF):
    """Font metrics information."""

    def elided_text(
        self,
        text: str,
        mode: constants.TextElideModeStr | constants.TextElideMode,
        width: float,
        flags=0,
    ) -> str:
        val = constants.TEXT_ELIDE_MODE.get_enum_value(mode)
        return self.elidedText(text, val, width, flags)

    def get_bounding_rect(self, *args) -> core.RectF:
        return core.RectF(self.boundingRect(*args))

    def get_tight_bounding_rect(self, text: str) -> core.RectF:
        return core.RectF(self.tightBoundingRect(text))



"""Object browser GUI in Qt."""

from __future__ import annotations

import logging

from prettyqt import constants, core, custom_widgets, gui, itemmodels, widgets
from prettyqt.itemmodels.pythonobjecttreemodel import (
    CommentsColumn,
    DocStringColumn,
    FileColumn,
    PathColumn,
    PrettyPrintColumn,
    ReprColumn,
    SourceCodeColumn,
    StrColumn,
)


DEFAULT_ATTR_DETAILS = [
    PathColumn,
    StrColumn,
    ReprColumn,
    PrettyPrintColumn,
    DocStringColumn,
    CommentsColumn,
    FileColumn,
    SourceCodeColumn,
]

logger = logging.getLogger(__name__)


class ObjectBrowser(widgets.MainWindow):
    """Info window for python objects."""

    def __init__(self, obj, stack=None, name: str = ""):
        super().__init__()
        self.set_title("Object browser")
        self.set_icon("mdi.language-python")
        self._auto_refresh = False
        self._refresh_rate = 2
        self._tree_model = itemmodels.PythonObjectTreeModel(obj)
        self._attr_details = [
            Klass(model=self._tree_model) for Klass in DEFAULT_ATTR_DETAILS
        ]

        # Toggle auto-refresh on/off
        self.toggle_auto_refresh_action = gui.Action(
            text="Auto-refresh",
            parent=self,
            checkable=True,
            status_tip=f"Auto refresh every {self._refresh_rate} seconds",
        )
        self.toggle_auto_refresh_action.toggled.connect(self.toggle_auto_refresh)

        # Add another refresh action with a different shortcut. An action must be added to
        # a visible widget for it to receive events. It is added to the main windows to
        # prevent it from being displayed again in the menu

        self.refresh_action_f5 = self.add_action(
            text="&Refresh2", shortcut="F5", triggered=self._tree_model.refresh_tree
        )
        self.central_splitter = widgets.Splitter(
            parent=self, orientation=constants.VERTICAL
        )
        self.setCentralWidget(self.central_splitter)

        # Tree widget
        self.obj_tree = widgets.TreeView(
            root_is_decorated=True, selection_behavior="rows"
        )
        self.obj_tree.set_model(self._tree_model)
        self.obj_tree.h_header = custom_widgets.FilterHeader(self.obj_tree)

        # Stretch last column?
        # It doesn't play nice when columns are hidden and then shown again.
        self.obj_tree.h_header.setObjectName("table_header")
        self.obj_tree.h_header.setSectionsMovable(True)
        self.obj_tree.h_header.setStretchLastSection(False)
        self.central_splitter.addWidget(self.obj_tree)

        # Bottom pane
        bottom_pane_widget = widgets.Widget()
        bottom_pane_widget.set_layout("horizontal", spacing=0, margin=5)
        self.central_splitter.addWidget(bottom_pane_widget)

        group_box = widgets.GroupBox("Details")
        bottom_pane_widget.box.addWidget(group_box)

        group_box.set_layout("horizontal", margin=2)

        # Radio buttons
        radio_widget = widgets.Widget()
        radio_widget.set_layout("vertical", margin=0)

        self.button_group = widgets.ButtonGroup(
            self, button_clicked=self._change_details_field
        )
        for button_id, attr_detail in enumerate(self._attr_details):
            radio_button = widgets.RadioButton(attr_detail.name)
            radio_widget.box.addWidget(radio_button)
            self.button_group.addButton(radio_button, button_id)

        self.button_group.button(0).setChecked(True)

        radio_widget.box.addStretch(1)
        group_box.box.addWidget(radio_widget)

        self.editor = widgets.PlainTextEdit(read_only=True, font=gui.Font.mono())
        group_box.box.addWidget(self.editor)

        # Splitter parameters
        self.central_splitter.setCollapsible(0, False)
        self.central_splitter.setCollapsible(1, True)
        self.central_splitter.setSizes([400, 200])
        self.central_splitter.setStretchFactor(0, 10)
        self.central_splitter.setStretchFactor(1, 0)

        selection_model = self.obj_tree.selectionModel()
        selection_model.currentRowChanged.connect(self._update_details)
        menubar = self.menuBar()
        file_menu = menubar.add_menu("&File")
        file_menu.add_action(text="C&lose", triggered=self.close, shortcut="Ctrl+W")
        file_menu.add_action(
            text="E&xit",
            triggered=lambda: widgets.app().closeAllWindows(),
            shortcut="Ctrl+Q",
        )

        view_menu = menubar.add_menu("&View")
        view_menu.add_action(
            text="&Refresh", triggered=self._tree_model.refresh_tree, shortcut="Ctrl+R"
        )
        view_menu.addAction(self.toggle_auto_refresh_action)

        view_menu.addSeparator()
        self.show_cols_submenu = self.obj_tree.h_header.createPopupMenu()
        self.show_cols_submenu.setTitle("Table columns")
        view_menu.add_menu(self.show_cols_submenu)
        view_menu.addSeparator()

        assert self._refresh_rate > 0
        self._refresh_timer = core.Timer(self)
        self._refresh_timer.setInterval(self._refresh_rate * 1000)
        self._refresh_timer.timeout.connect(self._tree_model.refresh_tree)
        self.toggle_auto_refresh_action.setChecked(self._auto_refresh)

    # @classmethod
    # def setup_example(cls):
    #     return cls(obj=["a", "b", "c"])

    @core.Slot(core.ModelIndex, core.ModelIndex)
    def _update_details(
        self, current_index: core.ModelIndex, _previous_index: core.ModelIndex
    ):
        """Show the object details in the editor given an index."""
        if self.obj_tree.model() is None or not current_index.isValid():
            return
        role = self.obj_tree.get_model(skip_proxies=True).ExtraRoles.TreeItemRole
        tree_item = current_index.data(role)
        self._update_details_for_item(tree_item)

    def _change_details_field(self, _button_id=None):
        """Change the field that is displayed in the details pane."""
        # logger.debug("_change_details_field: {}".format(_button_id))
        if self.obj_tree.model() is None:
            return
        current_index = self.obj_tree.selectionModel().currentIndex()
        if not current_index.isValid():
            return
        role = self.obj_tree.get_model(skip_proxies=True).ExtraRoles.TreeItemRole
        tree_item = current_index.data(role)
        self._update_details_for_item(tree_item)

    def _update_details_for_item(self, tree_item):
        """Show the object details in the editor given an tree_item."""
        button_id = self.button_group.checkedId()
        attr_details = self._attr_details[button_id]
        data = attr_details.get_data(tree_item)
        self.editor.set_text(data)
        self.editor.set_word_wrap_mode(attr_details.line_wrap)

    def toggle_auto_refresh(self, checked):
        """Toggle auto-refresh on/off."""
        if checked:
            logger.info("Auto-refresh on. Rate %g seconds", self._refresh_rate)
            self._refresh_timer.start()
        else:
            logger.info("Auto-refresh off")
            self._refresh_timer.stop()
        self._auto_refresh = checked

    def closeEvent(self, event):
        """Called when the window is closed."""
        self._refresh_timer.stop()
        self._refresh_timer.timeout.disconnect(self._tree_model.refresh_tree)
        self.toggle_auto_refresh_action.toggled.disconnect(self.toggle_auto_refresh)
        self.refresh_action_f5.triggered.disconnect(self._tree_model.refresh_tree)
        self.button_group.buttonClicked.disconnect(self._change_details_field)
        selection_model = self.obj_tree.selectionModel()
        selection_model.currentRowChanged.disconnect(self._update_details)
        self.close()
        event.accept()


if __name__ == "__main__":
    struct = dict(a={1, 2, frozenset([1, 2])})
    app = widgets.app()
    with app.debug_mode():
        object_browser = ObjectBrowser(struct)
        object_browser.show()
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtSpatialAudio
from prettyqt.utils import bidict, datatypes


MaterialStr = Literal[
    "transparent",
    "acoustic_ceiling_tiles",
    "brick_bare",
    "brick_painted",
    "concrete_block_coarse",
    "concrete_block_painted",
    "curtain_heavy",
    "fiber_glass_insulation",
    "glass_thin",
    "glass_thick",
    "grass",
    "linoleumon_concrete",
    "marble",
    "metal",
    "parquet_on_concrete",
    "plaste_rrough",
    "plaste_rsmooth",
    "plywood_panel",
    "polished_concrete_or_tile",
    "sheet_rock",
    "water_or_ice_surface",
    "wood_ceiling",
    "wood_panel",
    "uniform_material",
]

MATERIAL: bidict[MaterialStr, QtSpatialAudio.QAudioRoom.Material] = bidict(
    transparent=QtSpatialAudio.QAudioRoom.Material.Transparent,
    acoustic_ceiling_tiles=QtSpatialAudio.QAudioRoom.Material.AcousticCeilingTiles,
    brick_bare=QtSpatialAudio.QAudioRoom.Material.BrickBare,
    brick_painted=QtSpatialAudio.QAudioRoom.Material.BrickPainted,
    concrete_block_coarse=QtSpatialAudio.QAudioRoom.Material.ConcreteBlockCoarse,
    concrete_block_painted=QtSpatialAudio.QAudioRoom.Material.ConcreteBlockPainted,
    curtain_heavy=QtSpatialAudio.QAudioRoom.Material.CurtainHeavy,
    fiber_glass_insulation=QtSpatialAudio.QAudioRoom.Material.FiberGlassInsulation,
    glass_thin=QtSpatialAudio.QAudioRoom.Material.GlassThin,
    glass_thick=QtSpatialAudio.QAudioRoom.Material.GlassThick,
    grass=QtSpatialAudio.QAudioRoom.Material.Grass,
    linoleumon_concrete=QtSpatialAudio.QAudioRoom.Material.LinoleumOnConcrete,
    marble=QtSpatialAudio.QAudioRoom.Material.Marble,
    metal=QtSpatialAudio.QAudioRoom.Material.Metal,
    parquet_on_concrete=QtSpatialAudio.QAudioRoom.Material.ParquetOnConcrete,
    plaste_rrough=QtSpatialAudio.QAudioRoom.Material.PlasterRough,
    plaste_rsmooth=QtSpatialAudio.QAudioRoom.Material.PlasterSmooth,
    plywood_panel=QtSpatialAudio.QAudioRoom.Material.PlywoodPanel,
    polished_concrete_or_tile=QtSpatialAudio.QAudioRoom.Material.PolishedConcreteOrTile,
    sheet_rock=QtSpatialAudio.QAudioRoom.Material.Sheetrock,
    water_or_ice_surface=QtSpatialAudio.QAudioRoom.Material.WaterOrIceSurface,
    wood_ceiling=QtSpatialAudio.QAudioRoom.Material.WoodCeiling,
    wood_panel=QtSpatialAudio.QAudioRoom.Material.WoodPanel,
    uniform_material=QtSpatialAudio.QAudioRoom.Material.UniformMaterial,
)

WallStr = Literal[
    "left_wall",
    "right_wall",
    "floor",
    "ceiling",
    "front_wall",
    "back_wall",
]

WALL: bidict[WallStr, QtSpatialAudio.QAudioRoom.Wall] = bidict(
    left_wall=QtSpatialAudio.QAudioRoom.Wall.LeftWall,
    right_wall=QtSpatialAudio.QAudioRoom.Wall.RightWall,
    floor=QtSpatialAudio.QAudioRoom.Wall.Floor,
    ceiling=QtSpatialAudio.QAudioRoom.Wall.Ceiling,
    front_wall=QtSpatialAudio.QAudioRoom.Wall.FrontWall,
    back_wall=QtSpatialAudio.QAudioRoom.Wall.BackWall,
)


class AudioRoom(core.ObjectMixin, QtSpatialAudio.QAudioRoom):
    """Represents a room with its acoustic properties."""

    def set_wall_material(
        self,
        wall: WallStr | QtSpatialAudio.QAudioRoom.Wall,
        material: MaterialStr | QtSpatialAudio.QAudioRoom.Material,
    ):
        """Set the wall material.

        Args:
            wall: Wall to set
            material: material type
        """
        self.setWallMaterial(WALL.get_enum_value(wall), MATERIAL.get_enum_value(material))

    def get_wall_material(
        self, wall: WallStr | QtSpatialAudio.QAudioRoom.Wall
    ) -> MaterialStr:
        """Return wall material.

        Arguments:
            wall: wall to get material for
        """
        return MATERIAL.inverse[self.wallMaterial(WALL.get_enum_value(wall))]

    def set_dimensions(self, dimensions: datatypes.Vector3DType):
        self.setDimensions(datatypes.to_vector3d(dimensions))

    def set_position(self, position: datatypes.Vector3DType):
        self.setPosition(datatypes.to_vector3d(position))



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtCore
from prettyqt.utils import bidict


AlgorithmStr = Literal[
    "md_4",
    "md_5",
    "sha_1",
    "sha_224",
    "sha_256",
    "sha_384",
    "sha_512",
    "sha3_224",
    "sha3_256",
    "sha3_384",
    "sha3_512",
    "keccak_224",
    "keccak_256",
    "keccak_384",
    "keccak_512",
]


ALGORITHM: bidict[AlgorithmStr, QtCore.QCryptographicHash.Algorithm] = bidict(
    md_4=QtCore.QCryptographicHash.Algorithm.Md4,
    md_5=QtCore.QCryptographicHash.Algorithm.Md5,
    sha_1=QtCore.QCryptographicHash.Algorithm.Sha1,
    sha_224=QtCore.QCryptographicHash.Algorithm.Sha224,
    sha_256=QtCore.QCryptographicHash.Algorithm.Sha256,
    sha_384=QtCore.QCryptographicHash.Algorithm.Sha384,
    sha_512=QtCore.QCryptographicHash.Algorithm.Sha512,
    sha3_224=QtCore.QCryptographicHash.Algorithm.Sha3_224,
    sha3_256=QtCore.QCryptographicHash.Algorithm.Sha3_256,
    sha3_384=QtCore.QCryptographicHash.Algorithm.Sha3_384,
    sha3_512=QtCore.QCryptographicHash.Algorithm.Sha3_512,
    keccak_224=QtCore.QCryptographicHash.Algorithm.Keccak_224,
    keccak_256=QtCore.QCryptographicHash.Algorithm.Keccak_256,
    keccak_384=QtCore.QCryptographicHash.Algorithm.Keccak_384,
    keccak_512=QtCore.QCryptographicHash.Algorithm.Keccak_512,
)


class CryptographicHash(QtCore.QCryptographicHash):
    """Way to generate cryptographic hashes."""

    def __init__(self, method: QtCore.QCryptographicHash.Algorithm | AlgorithmStr):
        if isinstance(method, QtCore.QCryptographicHash.Algorithm):
            arg = method
        else:
            arg = ALGORITHM[method]
        super().__init__(arg)

    # def __str__(self):
    #     return bytes(self.result()).decode()

    def __bytes__(self):
        return self.get_result()

    def get_result(self) -> bytes:
        return bytes(self.result())

    @staticmethod
    def get_hash_length(
        method: AlgorithmStr | QtCore.QCryptographicHash.Algorithm,
    ) -> int:
        return CryptographicHash.hashLength(ALGORITHM.get_enum_value(method))



from __future__ import annotations

import logging

from prettyqt import constants, core, gui, widgets


logger = logging.getLogger(__name__)


class OrientedTableView(widgets.TableView):
    """TableView class with some convenience methods for oriented tables."""

    def __init__(
        self, orientation: constants.Orientation | constants.OrientationStr, **kwargs
    ):
        super().__init__(**kwargs)
        self.orientation = constants.ORIENTATION.get_enum_value(orientation)

    @classmethod
    def setup_example(cls):
        return cls(orientation="vertical")

    def get_higher_levels(self, levels: int) -> core.ItemSelection:
        model = self.model()
        if self.is_horizontal():
            # Get the header's selected columns
            # Removes the higher levels so that only the lowest level of the header
            # affects the data table selection
            return core.ItemSelection(
                model.index(0, 0), model.index(levels - 2, model.columnCount() - 1)
            )
        return core.ItemSelection(
            model.index(0, 0), model.index(model.rowCount() - 1, levels - 2)
        )

    def set_section_span(self, row: int, column: int, count: int):
        if self.is_horizontal():
            self.setSpan(row, column, 1, count)
        else:
            self.setSpan(column, row, count, 1)

    def sectionAt(self, val: int):
        return self.columnAt(val) if self.is_horizontal() else self.rowAt(val)

    def over_header_edge(self, position: int, margin: int = 3) -> int:
        # Return the index of the column this x position is on the right edge of
        left = self.sectionAt(position - margin)
        right = self.sectionAt(position + margin)
        if left != right != 0:
            # We're at the left edge of the first column
            return left
        return None

    def is_horizontal(self) -> bool:
        return self.orientation == constants.HORIZONTAL

    def sectionWidth(self, val: int):
        return self.columnWidth(val) if self.is_horizontal() else self.rowHeight(val)

    def setSectionWidth(self, section: int, width: int):
        if self.is_horizontal():
            self.setColumnWidth(section, width)
        else:
            self.setRowHeight(section, width)

    def get_split_cursor(self) -> gui.Cursor:
        if self.is_horizontal():
            return gui.Cursor(constants.CursorShape.SplitHCursor)
        return gui.Cursor(constants.CursorShape.SplitVCursor)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    with app.debug_mode():
        app.set_style("fusion")
        view = OrientedTableView()
        view.show()
        app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core, widgets


if TYPE_CHECKING:
    from collections.abc import Sequence


class IntLineEdit(widgets.LineEdit):
    value_changed = core.Signal(int)

    def __init__(self, *args, object_name: str = "int_lineedit", **kwargs):
        super().__init__(*args, object_name=object_name, **kwargs)
        self.set_validator("integer", bottom=0, top=1000000000)

    def get_value(self) -> int:
        val = super().get_value()
        return int(val) if val and val.isnumeric() else 0

    def set_value(self, value: int | str):
        super().set_value(str(value))

    value = core.Property(
        int,
        get_value,
        set_value,
        doc="Current Value as integer",
    )


class FloatLineEdit(widgets.LineEdit):
    value_changed = core.Signal(float)

    def __init__(self, *args, object_name: str = "float_lineedit", **kwargs):
        super().__init__(*args, object_name=object_name, **kwargs)
        self.set_validator("double", bottom=0.0)

    def get_value(self) -> float:
        val = super().get_value()
        try:
            return float(val)
        except ValueError:
            return 0.0

    def set_value(self, value: float | str):
        super().set_value(str(value))

    value = core.Property(
        float,
        get_value,
        set_value,
        doc="Current Value as float",
    )


class UrlLineEdit(widgets.LineEdit):
    value_changed = core.Signal(core.QUrl)

    def __init__(self, *args, object_name: str = "float_lineedit", **kwargs):
        super().__init__(*args, object_name=object_name, **kwargs)
        self.set_validator("website")

    def get_value(self) -> core.QUrl:
        val = super().get_value()
        return core.QUrl.fromUserInput(val)

    def set_value(self, value: core.QUrl | str):
        super().set_value(value.toString() if isinstance(value, core.QUrl) else value)

    value = core.Property(
        core.QUrl,
        get_value,
        set_value,
        doc="Current Value as QUrl",
    )


class StringListEdit(widgets.LineEdit):
    value_changed = core.Signal(list)

    def __init__(self, *args, object_name: str = "str_list_lineedit", **kwargs):
        super().__init__(*args, object_name=object_name, **kwargs)

    def _on_value_change(self):
        value = self.get_value()
        self.value_changed.emit(value)

    def get_value(self) -> list[str]:
        val = super().get_value()
        return val.split(",")

    def set_value(self, value: Sequence[str]):
        super().set_value(",".join(value))

    value = core.Property(
        list,
        get_value,
        set_value,
        doc="Current Value as list",
    )


if __name__ == "__main__":
    app = widgets.app()
    widget = FloatLineEdit(window_title="Test")
    widget.set_value("")
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import gui


class PaintDeviceWindowMixin(gui.WindowMixin, gui.PaintDeviceMixin):
    pass


class PaintDeviceWindow(PaintDeviceWindowMixin, gui.QPaintDeviceWindow):
    pass


if __name__ == "__main__":
    window = PaintDeviceWindow()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtQuick


class QuickImageResponse(core.ObjectMixin, QtQuick.QQuickImageResponse):
    """Interface for asynchronous image loading in QQuickAsyncImageProvider."""


if __name__ == "__main__":
    item = QuickImageResponse()



from __future__ import annotations

import logging
import pathlib
import sys

import prettyqt
from prettyqt import prettyqtmarkdown
import mknodes as mk


logger = logging.getLogger(__name__)
logging.basicConfig(stream=sys.stdout, level=logging.INFO)


def is_qt_module(module):
    return hasattr(module, "QT_MODULE")


def build_root(project: mk.Project):
    old = project.linkprovider
    provider = prettyqtmarkdown.QtLinkProvider(
        old.base_url, old.use_directory_urls, include_stdlib=True
    )
    provider.add_inv_file("docs/qt6.inv", "https://doc.qt.io/qtforpython/")
    project.linkprovider = provider
    project.theme.announcement_bar = mk.MkMetadataBadges("dependencies")
    nav_file = pathlib.Path(__file__).parent / "SUMMARY.md"
    root_nav = mk.MkNav()
    root_nav.parse.file(nav_file)
    project._root = root_nav
    root_nav.associated_project = project
    qt_docs = root_nav.add_doc(prettyqt, section_name="Qt-based modules")
    # simple/qt
    extra_docs = root_nav.add_doc(prettyqt, section_name="Additional modules")
    # octicons/plus-16
    populate_docs(qt_docs, is_qt_module)
    populate_docs(extra_docs, lambda x: not is_qt_module(x))

    dev_nav = root_nav.add_nav("Development")
    populate_dev_section(dev_nav)
    return root_nav


def populate_docs(doc_nav: mk.MkDoc, predicate):
    prettyqt.import_all()
    for submod in doc_nav.iter_modules(predicate=predicate):
        doc_nav.add_doc(submod, class_template="classpage_custom.jinja", flatten_nav=True)


def populate_dev_section(nav: mk.MkNav):
    page = nav.add_page("Changelog")
    page += mk.MkChangelog()

    page = nav.add_page("Dependencies", hide="toc")
    page += mk.MkDependencyTable()

    page = nav.add_page("Code of conduct")
    page += mk.MkCodeOfConduct()

    page = nav.add_page("Contributing")
    page += mk.MkCommitConventions()
    page += mk.MkPullRequestGuidelines()

    page = nav.add_page("Setting up the environment")
    page += mk.MkDevEnvSetup()

    node = mk.MkLicense()
    page = nav.add_page("License", hide="toc")
    page += node


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    nav_file = pathlib.Path(__file__).parent / "SUMMARY.md"
    root_nav = mk.MkNav()
    root_nav.parse.file(nav_file)
    table = prettyqtmarkdown.MarkdownWidget()
    table.set_markdown(root_nav)
    table.show()
    app.exec()



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Literal, Self

from prettyqt import core
from prettyqt.qt import QtGui
from prettyqt.utils import bidict, get_repr


if TYPE_CHECKING:
    from prettyqt import constants


SEQUENCE_MATCHES = bidict(
    none=QtGui.QKeySequence.SequenceMatch.NoMatch,
    partial=QtGui.QKeySequence.SequenceMatch.PartialMatch,
    exact=QtGui.QKeySequence.SequenceMatch.ExactMatch,
)

SequenceMatchStr = Literal["none", "partial", "exact"]

SEQUENCE_FORMAT = bidict(
    native=QtGui.QKeySequence.SequenceFormat.NativeText,
    portable=QtGui.QKeySequence.SequenceFormat.PortableText,
)

SequenceFormatStr = Literal["native", "Portable"]

STANDARD_KEYS = bidict(
    add_tab=QtGui.QKeySequence.StandardKey.AddTab,
    back=QtGui.QKeySequence.StandardKey.Back,
    backspace=QtGui.QKeySequence.StandardKey.Backspace,
    bold=QtGui.QKeySequence.StandardKey.Bold,
    close=QtGui.QKeySequence.StandardKey.Close,
    copy=QtGui.QKeySequence.StandardKey.Copy,
    cut=QtGui.QKeySequence.StandardKey.Cut,
    delete=QtGui.QKeySequence.StandardKey.Delete,
    delete_end_of_line=QtGui.QKeySequence.StandardKey.DeleteEndOfLine,
    delete_end_of_word=QtGui.QKeySequence.StandardKey.DeleteEndOfWord,
    delete_start_of_word=QtGui.QKeySequence.StandardKey.DeleteStartOfWord,
    delete_complete_line=QtGui.QKeySequence.StandardKey.DeleteCompleteLine,
    find=QtGui.QKeySequence.StandardKey.Find,
    find_next=QtGui.QKeySequence.StandardKey.FindNext,
    find_previous=QtGui.QKeySequence.StandardKey.FindPrevious,
    forward=QtGui.QKeySequence.StandardKey.Forward,
    help_contents=QtGui.QKeySequence.StandardKey.HelpContents,
    instert_line_separator=QtGui.QKeySequence.StandardKey.InsertLineSeparator,
    insert_paragraph_separator=QtGui.QKeySequence.StandardKey.InsertParagraphSeparator,
    italic=QtGui.QKeySequence.StandardKey.Italic,
    move_to_end_of_block=QtGui.QKeySequence.StandardKey.MoveToEndOfBlock,
    move_to_end_of_document=QtGui.QKeySequence.StandardKey.MoveToEndOfDocument,
    move_to_end_of_line=QtGui.QKeySequence.StandardKey.MoveToEndOfLine,
    move_to_next_char=QtGui.QKeySequence.StandardKey.MoveToNextChar,
    move_to_next_line=QtGui.QKeySequence.StandardKey.MoveToNextLine,
    move_to_next_page=QtGui.QKeySequence.StandardKey.MoveToNextPage,
    move_to_next_word=QtGui.QKeySequence.StandardKey.MoveToNextWord,
    move_to_previous_char=QtGui.QKeySequence.StandardKey.MoveToPreviousChar,
    move_to_previous_line=QtGui.QKeySequence.StandardKey.MoveToPreviousLine,
    move_to_previous_page=QtGui.QKeySequence.StandardKey.MoveToPreviousPage,
    move_to_previous_word=QtGui.QKeySequence.StandardKey.MoveToPreviousWord,
    move_to_start_of_block=QtGui.QKeySequence.StandardKey.MoveToStartOfBlock,
    move_to_start_of_document=QtGui.QKeySequence.StandardKey.MoveToStartOfDocument,
    move_to_start_of_line=QtGui.QKeySequence.StandardKey.MoveToStartOfLine,
    new=QtGui.QKeySequence.StandardKey.New,
    next_child=QtGui.QKeySequence.StandardKey.NextChild,
    open=QtGui.QKeySequence.StandardKey.Open,
    paste=QtGui.QKeySequence.StandardKey.Paste,
    preferences=QtGui.QKeySequence.StandardKey.Preferences,
    previous_child=QtGui.QKeySequence.StandardKey.PreviousChild,
    print=QtGui.QKeySequence.StandardKey.Print,
    quit=QtGui.QKeySequence.StandardKey.Quit,
    redo=QtGui.QKeySequence.StandardKey.Redo,
    refresh=QtGui.QKeySequence.StandardKey.Refresh,
    replace=QtGui.QKeySequence.StandardKey.Replace,
    save_as=QtGui.QKeySequence.StandardKey.SaveAs,
    save=QtGui.QKeySequence.StandardKey.Save,
    select_all=QtGui.QKeySequence.StandardKey.SelectAll,
    deselect=QtGui.QKeySequence.StandardKey.Deselect,
    select_end_of_block=QtGui.QKeySequence.StandardKey.SelectEndOfBlock,
    select_end_of_document=QtGui.QKeySequence.StandardKey.SelectEndOfDocument,
    select_end_of_line=QtGui.QKeySequence.StandardKey.SelectEndOfLine,
    select_next_char=QtGui.QKeySequence.StandardKey.SelectNextChar,
    select_next_line=QtGui.QKeySequence.StandardKey.SelectNextLine,
    select_next_page=QtGui.QKeySequence.StandardKey.SelectNextPage,
    select_next_word=QtGui.QKeySequence.StandardKey.SelectNextWord,
    select_previous_char=QtGui.QKeySequence.StandardKey.SelectPreviousChar,
    select_previous_line=QtGui.QKeySequence.StandardKey.SelectPreviousLine,
    select_previous_page=QtGui.QKeySequence.StandardKey.SelectPreviousPage,
    select_previous_word=QtGui.QKeySequence.StandardKey.SelectPreviousWord,
    select_start_of_block=QtGui.QKeySequence.StandardKey.SelectStartOfBlock,
    select_start_of_document=QtGui.QKeySequence.StandardKey.SelectStartOfDocument,
    select_start_of_line=QtGui.QKeySequence.StandardKey.SelectStartOfLine,
    underline=QtGui.QKeySequence.StandardKey.Underline,
    undo=QtGui.QKeySequence.StandardKey.Undo,
    unknown_key=QtGui.QKeySequence.StandardKey.UnknownKey,
    whats_this=QtGui.QKeySequence.StandardKey.WhatsThis,
    zoom_in=QtGui.QKeySequence.StandardKey.ZoomIn,
    zoom_out=QtGui.QKeySequence.StandardKey.ZoomOut,
    full_screen=QtGui.QKeySequence.StandardKey.FullScreen,
    cancel=QtGui.QKeySequence.StandardKey.Cancel,
)


class KeySequence(QtGui.QKeySequence):
    """Encapsulates a key sequence as used by shortcuts."""

    def __init__(self, *args, **kwargs):
        match args:
            case (str(),) if args[0] in STANDARD_KEYS:
                super().__init__(STANDARD_KEYS[args[0]])
            case _:
                super().__init__(*args, **kwargs)

    def __str__(self):
        return self.toString()

    def __repr__(self):
        return get_repr(self, self.toString())

    def __bool__(self):
        return not self.isEmpty()

    def __reduce__(self):
        return type(self), (self.toString(),)

    def __format__(self, format_spec: SequenceFormatStr):
        if format_spec in SEQUENCE_FORMAT:
            return self.toString(SEQUENCE_FORMAT[format_spec])
        return super().__format__(format_spec)

    def __getitem__(self, item) -> core.KeyCombination:
        item = super().__getitem__(item)
        return core.KeyCombination(item)

    @property
    def _toString(self) -> str:
        return self.toString()

    __match_args__ = ("_toString",)

    def get_matches(self, seq: QtGui.QKeySequence | str) -> SequenceMatchStr:
        if isinstance(seq, str):
            seq = KeySequence(seq)
        return SEQUENCE_MATCHES.inverse[self.matches(seq)]

    @functools.singledispatchmethod
    @classmethod
    def to_shortcut_str(cls, key: constants.Key, mod: int = 0) -> str:
        return cls(mod | key).toString()

    @to_shortcut_str.register
    @classmethod
    def _(cls, key: core.QKeyCombination) -> str:
        return cls(key).toString()

    @classmethod
    def get_key_bindings(cls, button) -> list[Self]:
        return [cls(i) for i in cls.keyBindings(button)]


if __name__ == "__main__":
    seq = KeySequence("select_all")



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtGui
from prettyqt.utils import bidict


WORD_WRAP_MODE = bidict(
    none=QtGui.QTextOption.WrapMode.NoWrap,
    word=QtGui.QTextOption.WrapMode.WordWrap,
    manual=QtGui.QTextOption.WrapMode.ManualWrap,
    anywhere=QtGui.QTextOption.WrapMode.WrapAnywhere,
    boundary_or_anywhere=QtGui.QTextOption.WrapMode.WrapAtWordBoundaryOrAnywhere,
)

WordWrapModeStr = Literal["none", "word", "manual", "anywhere", "boundary_or_anywhere"]

TAB_TYPE = bidict(
    left=QtGui.QTextOption.TabType.LeftTab,
    right=QtGui.QTextOption.TabType.RightTab,
    center=QtGui.QTextOption.TabType.CenterTab,
    delimiter=QtGui.QTextOption.TabType.DelimiterTab,
)

TabTypeStr = Literal["left", "right", "center", "delimiter"]


FLAG = bidict(
    include_trailing_whitespaces=QtGui.QTextOption.Flag.IncludeTrailingSpaces,
    show_tabs_and_spaces=QtGui.QTextOption.Flag.ShowTabsAndSpaces,
    show_separators=QtGui.QTextOption.Flag.ShowLineAndParagraphSeparators,
    show_document_terminator=QtGui.QTextOption.Flag.ShowDocumentTerminator,
    add_space_for_separators=QtGui.QTextOption.Flag.AddSpaceForLineAndParagraphSeparators,
    suppress_colors=QtGui.QTextOption.Flag.SuppressColors,
)

FlagStr = Literal[
    "include_trailing_whitespaces",
    "show_tabs_and_spaces",
    "show_separators",
    "show_document_terminator",
    "add_space_for_separators",
    "suppress_colors",
]


class TextOption(QtGui.QTextOption):
    def set_flags(self, **flags):
        current = self.flags()
        for k, v in flags.items():
            if v:
                current |= FLAG[k]
            else:
                current &= ~FLAG[k]
        self.setFlags(current)
        # if show:
        #     self.setFlags(self.flags() | QtGui.QTextOption.ShowTabsAndSpaces)
        # else:
        #     self.setFlags(self.flags() & ~QtGui.QTextOption.ShowTabsAndSpaces)


if __name__ == "__main__":
    doc = TextOption()



from __future__ import annotations

from prettyqt import core, location, positioning


class Place(location.QPlace):
    """Represents a set of data about a place."""

    def __bool__(self):
        return not self.isEmpty()

    def __setitem__(self, index: str, val: location.QPlaceAttribute | None):
        if val is None:
            val = location.PlaceAttribute()
        self.setExtendedAttribute(index, val)

    def __getitem__(self, index: str) -> location.PlaceAttribute:
        attr = self.extendedAttribute(index)
        return location.PlaceAttribute(attr)

    def get_categories(self) -> list[location.PlaceCategory]:
        return [location.PlaceCategory(i) for i in self.categories()]

    def get_contact_details(self, contact_type: str) -> list[location.PlaceContactDetail]:
        return [location.PlaceContactDetail(i) for i in self.contactDetails(contact_type)]

    def set_content(self, typ: str, value: dict[int, location.QPlaceContent]):
        self.setContent(location.placecontent.TYPE.get_enum_value(typ), value)

    def get_content(self, typ: str) -> dict[int, location.QPlaceContent]:
        return self.content(location.placecontent.TYPE.get_enum_value(typ))

    def get_icon(self) -> location.PlaceIcon | None:
        icon = self.icon()
        return None if icon.isEmpty() else location.PlaceIcon(icon)

    def get_location(self) -> positioning.GeoLocation:
        return positioning.GeoLocation(self.location())

    def get_primary_website(self) -> core.Url:
        return core.Url(self.primaryWebsite())

    def get_ratings(self) -> location.PlaceRatings:
        return location.PlaceRatings(self.ratings())

    def get_supplier(self) -> location.PlaceSupplier:
        return location.PlaceSupplier(self.supplier())

    def get_visibility(self) -> location.VisibilityStr:
        """Return visibility.

        Returns:
            Visibility
        """
        return location.VISIBILITY.inverse[self.visibility()]


if __name__ == "__main__":
    place = Place()
    attr = location.PlaceAttribute()
    place["test"] = attr
    assert place["test"] == attr



from __future__ import annotations

from prettyqt import constants, gui


class TextTableFormat(gui.textframeformat.TextFrameFormatMixin, gui.QTextTableFormat):
    def __bool__(self):
        return self.isValid()

    def set_alignment(self, alignment: constants.AlignmentStr | constants.AlignmentFlag):
        """Set the alignment of the format.

        Args:
            alignment: alignment for the format
        """
        self.setAlignment(constants.ALIGNMENTS.get_enum_value(alignment))

    def get_alignment(self) -> constants.AlignmentStr:
        """Return current alignment.

        Returns:
            alignment
        """
        return constants.ALIGNMENTS.inverse[self.alignment()]

    def get_column_width_constraints(self) -> list[gui.TextLength]:
        return [gui.TextLength(i) for i in self.columnWidthConstraints()]


if __name__ == "__main__":
    fmt = TextTableFormat()
    print(fmt)



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core, gui, widgets
from prettyqt.utils import get_repr


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class KeyCombinationEdit(widgets.WidgetMixin, widgets.QKeySequenceEdit):
    value_changed = core.Signal(core.QKeyCombination)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setMaximumSequenceLength(1)

    def __repr__(self):
        return get_repr(self, self.get_value())

    def _on_value_change(self, val):
        self.value_changed.emit(val[0])

    def set_value(self, value: datatypes.KeyCombinationType):
        if not isinstance(value, gui.QKeySequence):
            value = gui.KeySequence(value)
        if len(value) == 0:
            self.clear()
        else:
            self.setKeySequence(gui.KeySequence(value[0]))

    def get_value(self) -> str:
        seq = self.keySequence()
        return seq[0].toString() if seq.count() > 0 else ""

    def is_valid(self) -> bool:
        return True


if __name__ == "__main__":
    app = widgets.app()
    widget = KeyCombinationEdit()
    widget.set_value("")
    widget.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from prettyqt import core, gui, widgets
from prettyqt.utils import bidict, colors, texteditselecter


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


AutoFormattingStr = Literal["none", "bullet_list", "all"]

AUTO_FORMATTING: bidict[AutoFormattingStr, widgets.QTextEdit.AutoFormattingFlag] = bidict(
    none=widgets.QTextEdit.AutoFormattingFlag.AutoNone,
    bullet_list=widgets.QTextEdit.AutoFormattingFlag.AutoBulletList,
    all=widgets.QTextEdit.AutoFormattingFlag.AutoAll,
)

LineWrapModeStr = Literal[
    "none", "widget_width", "fixed_pixel_width", "fixed_column_width"
]

LINE_WRAP_MODE: bidict[LineWrapModeStr, widgets.QTextEdit.LineWrapMode] = bidict(
    none=widgets.QTextEdit.LineWrapMode.NoWrap,
    widget_width=widgets.QTextEdit.LineWrapMode.WidgetWidth,
    fixed_pixel_width=widgets.QTextEdit.LineWrapMode.FixedPixelWidth,
    fixed_column_width=widgets.QTextEdit.LineWrapMode.FixedColumnWidth,
)


class TextEditMixin(widgets.AbstractScrollAreaMixin):
    value_changed = core.Signal(str)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.textChanged.connect(self.on_value_change)
        self.selecter = texteditselecter.TextEditSelecter(self)

    def __add__(self, other: str) -> TextEdit:
        self.append_text(other)
        return self

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "autoFormatting": AUTO_FORMATTING,
            "lineWrapMode": LINE_WRAP_MODE,
            "wordWrapMode": gui.textoption.WORD_WRAP_MODE,
        }
        return maps

    def on_value_change(self) -> None:
        self.value_changed.emit(self.text())

    def set_text(self, text: str) -> None:
        self.setPlainText(text)

    def append_text(
        self,
        text: str,
        newline: bool = True,
        ensure_visible: Literal["always", "when_bottom", "never"] = "always",
    ):
        scrollbar = self.verticalScrollBar()
        at_bottom = scrollbar.value() >= (scrollbar.maximum() - 4)
        prev_val = scrollbar.value()
        if newline:
            self.append(text)
        else:
            self.selecter.move_cursor("end")
            self.insertHtml(text)
            self.selecter.move_cursor("end")
        match ensure_visible:
            case "always":
                self.ensureCursorVisible()
            case "when_bottom":
                if at_bottom:
                    self.ensureCursorVisible()
            case "never":
                scrollbar.setValue(prev_val)

    def text(self) -> str:
        return self.toPlainText()

    def set_read_only(self, value: bool = True) -> None:
        self.setReadOnly(value)

    def set_text_color(self, color: datatypes.ColorType) -> None:
        color = colors.get_color(color)
        self.setTextColor(color)

    def set_line_wrap_mode(self, mode: LineWrapModeStr | widgets.QTextEdit.LineWrapMode):
        """Set line wrap mode.

        Args:
            mode: line wrap mode to use
        """
        self.setLineWrapMode(LINE_WRAP_MODE.get_enum_value(mode))

    def get_line_wrap_mode(self) -> LineWrapModeStr:
        """Get the current wrap mode.

        Returns:
            Wrap mode
        """
        return LINE_WRAP_MODE.inverse[self.lineWrapMode()]

    def set_auto_formatting(
        self, mode: AutoFormattingStr | widgets.QTextEdit.AutoFormattingFlag
    ):
        """Set auto formatting mode.

        Args:
            mode: auto formatting mode to use
        """
        self.setAutoFormatting(AUTO_FORMATTING.get_enum_value(mode))

    def get_auto_formatting(self) -> AutoFormattingStr:
        """Get the current auto formatting mode.

        Returns:
            Auto formatting mode
        """
        return AUTO_FORMATTING.inverse[self.autoFormatting()]

    def set_word_wrap_mode(
        self, mode: gui.textoption.WordWrapModeStr | gui.QTextOption.WrapMode
    ):
        """Set word wrap mode.

        Args:
            mode: word wrap mode to use
        """
        self.setWordWrapMode(gui.textoption.WORD_WRAP_MODE.get_enum_value(mode))

    def get_word_wrap_mode(self) -> gui.textoption.WordWrapModeStr:
        """Get the current word wrap mode.

        Returns:
            Word wrap mode
        """
        return gui.textoption.WORD_WRAP_MODE.inverse[self.wordWrapMode()]


class TextEdit(TextEditMixin, widgets.QTextEdit):
    """Widget that is used to edit and display both plain and rich text."""


if __name__ == "__main__":
    app = widgets.app()
    widget = TextEdit("This is a test")
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import gui


class NotStrictValidator(gui.Validator):
    """Validator which turns Invalid state into Intermediate state."""

    ID = "not_strict"

    def __init__(self, validator: gui.QValidator | None = None, **kwargs):
        self._validator = validator
        super().__init__(**kwargs)

    def __eq__(self, other: object):
        return (
            isinstance(other, NotStrictValidator) and other._validator == self._validator
        )

    def validate(self, text: str, pos: int = 0) -> tuple[gui.QValidator.State, str, int]:
        state, text, pos = self._validator.validate(text, pos)
        is_invalid = state == self.State.Invalid
        return self.State.Intermediate if is_invalid else state, text, pos


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = widgets.LineEdit()
    widget.set_validator("alphanumeric", strict=False)
    widget.show()
    with app.debug_mode():
        app.exec()
    print(widget.hasAcceptableInput())



from __future__ import annotations

from prettyqt import constants, core, gui, widgets


class FlowLayout(widgets.Layout):
    """Layout which adjusts its children based on space available."""

    ID = "flow"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._items = []

    def __add__(self, other: widgets.QWidget | widgets.QLayout) -> FlowLayout:
        if not isinstance(other, widgets.QWidget | widgets.QLayout):
            raise TypeError
        self.add(other)
        return self

    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addLayout(self, layout: widgets.QLayout):
        widget = widgets.Widget()
        widget.setLayout(layout)
        self.addWidget(widget)

    def addItem(self, item):
        self._items.append(item)

    def itemAt(self, idx):
        try:
            return self._items[idx]
        except IndexError:
            pass

    def takeAt(self, idx):
        try:
            return self._items.pop(idx)
        except IndexError:
            pass

    def count(self):
        return len(self._items)

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        return self._do_layout(core.Rect(0, 0, width, 0), apply_geometry=False)

    def setGeometry(self, rect):
        super().setGeometry(rect)
        self._do_layout(rect, apply_geometry=True)

    def expandingDirections(self):
        return constants.Orientation(0)

    def minimumSize(self):
        size = core.QSize()
        for item in self._items:
            size = size.expandedTo(item.minimumSize())
        left, top, right, bottom = self.getContentsMargins()  # pyright: ignore
        return size + core.QSize(left + right, top + bottom)

    sizeHint = minimumSize  # noqa: N815

    def smart_spacing(self, horizontal: bool = True) -> int:
        p = self.parent()
        if p is None:
            return -1
        if p.isWidgetType():
            assert isinstance(p, widgets.QWidget)
            which = (
                widgets.Style.PixelMetric.PM_LayoutHorizontalSpacing
                if horizontal
                else widgets.Style.PixelMetric.PM_LayoutVerticalSpacing
            )
            return p.style().pixelMetric(which, None, p)
        return p.spacing()

    def _do_layout(self, rect: core.QRect, apply_geometry: bool = False) -> int:
        erect = rect.marginsRemoved(self.contentsMargins())
        x, y = erect.x(), erect.y()

        line_height = 0

        def layout_spacing(wid, horizontal: bool = True):
            if (ans := self.smart_spacing(horizontal)) != -1:
                return ans
            if wid is None:
                return 0
            return wid.style().layoutSpacing(
                widgets.SizePolicy.ControlType.PushButton,
                widgets.SizePolicy.ControlType.PushButton,
                constants.HORIZONTAL if horizontal else constants.VERTICAL,
            )

        lines = []
        current_line = []
        gmap = {}
        for item in self._items:
            isz, wid = item.sizeHint(), item.widget()
            hs, vs = layout_spacing(wid), layout_spacing(wid, False)

            next_x = x + isz.width() + hs
            if next_x - hs > erect.right() and line_height > 0:
                x = erect.x()
                y = y + line_height + vs
                next_x = x + isz.width() + hs
                lines.append((line_height, current_line))
                current_line = []
                line_height = 0
            if apply_geometry:
                gmap[item] = x, y, isz
            x = next_x
            line_height = max(line_height, isz.height())
            current_line.append((item, isz.height()))

        lines.append((line_height, current_line))

        if apply_geometry:
            for line_height, items in lines:
                for item, item_height in items:
                    x, wy, isz = gmap[item]
                    if item_height < line_height:
                        wy += (line_height - item_height) // 2
                    item.setGeometry(core.Rect(core.Point(x, wy), isz))

        return y + line_height - rect.y() + self.contentsMargins().bottom()


class Separator(widgets.Widget):
    """Vertical separator lines usable in FlowLayout."""

    def __init__(self, *args, widget_for_height: widgets.QWidget | None = None, **kwargs):
        """You must provide a widget in the layout either here or with setBuddy.

        The height of the separator is computed using this widget,
        """
        super().__init__(*args, **kwargs)
        self.bcol = widgets.app().palette().color(gui.Palette.ColorRole.Text)
        self.update_brush()
        self.widget_for_height = widget_for_height
        self.set_size_policy("fixed", "minimum_expanding")

    def update_brush(self):
        self.brush = gui.Brush(self.bcol)
        self.update()

    def setBuddy(self, widget_for_height: widgets.QWidget):
        """See __init__. This is repurposed to support Qt Designer .ui files."""
        self.widget_for_height = widget_for_height

    def sizeHint(self):
        return core.QSize(
            1, 1 if self.widget_for_height is None else self.widget_for_height.height()
        )

    def paintEvent(self, ev):
        with gui.Painter(self) as painter:
            # Purely subjective: shorten the line a bit to look 'better'
            r = ev.rect()
            r.setTop(r.top() + 3)
            r.setBottom(r.bottom() - 3)
            painter.fillRect(r, self.brush)


if __name__ == "__main__":
    app = widgets.app()
    widget = widgets.Widget()
    layout = FlowLayout()
    layout.add(widgets.PushButton("Short"))
    layout.add(widgets.PushButton("Longer"))
    layout.add(widgets.PushButton("Different text"))
    layout.add(widgets.PushButton("More text"))
    layout.add(widgets.PushButton("Even longer button text"))
    widget.set_layout(layout)
    widget.show()
    app.exec()



from __future__ import annotations

import functools
from operator import or_

from prettyqt import constants
from prettyqt.qt import QtCore
from prettyqt.utils import bidict, get_repr, serializemixin


Mod = constants.KeyboardModifier
Key = constants.Key

MODIFIERS = bidict({
    "None": Mod.NoModifier,
    "Shift": Mod.ShiftModifier,
    "Ctrl": Mod.ControlModifier,
    "Alt": Mod.AltModifier,
    "Meta": Mod.MetaModifier,
    "Ctrl+Shift": Mod.ControlModifier | Mod.ShiftModifier,
    "Ctrl+Alt": Mod.ControlModifier | Mod.AltModifier,
    "Alt+Shift": Mod.AltModifier | Mod.ShiftModifier,
})

MODIFIER_KEYS = frozenset({Key.Key_Shift, Key.Key_Control, Key.Key_Meta, Key.Key_Alt})
ARROW_KEYS = frozenset({Key.Key_Up, Key.Key_Down, Key.Key_Left, Key.Key_Right})
FUNC_ARROW_KEYS = frozenset({Key.Key_Home, Key.Key_End, Key.Key_PageUp, Key.Key_PageDown})

_SYMBOLS = {
    "Esc": "Escape",
    "!": "Exclam",
    '"': "QuoteDbl",
    "#": "NumberSign",
    "$": "Dollar",
    "%": "Percent",
    "&": "Ampersand",
    "'": "Apostrophe",
    "(": "ParenLeft",
    ")": "ParenRight",
    "*": "Asterisk",
    "+": "Plus",
    ",": "Comma",
    "-": "Minus",
    ".": "Period",
    "/": "Slash",
    ":": "Colon",
    ";": "Semicolon",
    "<": "Less",
    "=": "Equal",
    ">": "Greater",
    "?": "Question",
    "@": "At",
    "[": "BracketLeft",
    "\\": "Backslash",
    "]": "BracketRight",
    "^": "AsciiCircum",
    "_": "Underscore",
    "`": "QuoteLeft",
    "{": "BraceLeft",
    "|": "Bar",
    "}": "BraceRight",
    "~": "AsciiTilde",
}


class Keys:
    """Key extension."""

    No = Key(-1)
    Any = Key(-2)

    ALPHA = Key(913)  # α  # noqa: RUF003
    OMEGA = Key(937)  # ω

    CYR_A = Key(1040)  # а  # noqa: RUF003
    CYR_YA = Key(1071)  # я


class KeyCombination(serializemixin.SerializeMixin, QtCore.QKeyCombination):
    """Stores a combination of a key with optional modifiers."""

    def __init__(self, *args, **kwargs):
        match args:
            case (str(),) if args[0] in MODIFIERS.inverse.values():
                mods = args[0].split("+")
                qtmod = functools.reduce(or_, [MODIFIERS[m] for m in mods])
                super().__init__(qtmod, Keys.No)
                return
            case (str(),):
                *mods, btn = args[0].split("+")

                # get modifiler
                qtmod = (
                    functools.reduce(or_, [MODIFIERS[m] for m in mods])
                    if mods
                    else Mod.NoModifier
                )
                # get button
                if btn in _SYMBOLS:
                    btn = _SYMBOLS[btn]
                if btn.isalnum():
                    btn = btn.upper()
                qtkey = getattr(Key, f"Key_{btn}") if btn != "{}" else Keys.Any
                super().__init__(qtmod, qtkey)
            case (QtCore.QEvent(),):
                modifier = args[0].modifiers()
                modifier ^= Mod.KeypadModifier
                key = args[0].key()
                if key in MODIFIER_KEYS:  # modifier only
                    key = Keys.No
                super().__init__(key, modifier)
            case _:
                super().__init__(*args, **kwargs)

    def __eq__(self, other):
        if isinstance(other, str | Key):
            other = KeyCombination(other)
        return super().__eq__(other)

    def __add__(self, other):
        from prettyqt import gui

        return gui.KeySequence(self, other)

    def __repr__(self):
        return get_repr(self, self.key(), self.keyboardModifiers())

    def is_typing(self) -> bool:
        """True if key is a letter or number."""
        mod_ok = self.keyboardModifiers() in (Mod.NoModifier, Mod.ShiftModifier)
        key = self.key()
        key_ok = (
            Key.Key_Exclam <= key <= Key.Key_ydiaeresis
            or Keys.ALPHA <= key <= Keys.OMEGA
            or Keys.CYR_A <= key <= Keys.CYR_YA
        )
        return mod_ok and key_ok

    def is_moving(self) -> bool:
        """True if arrows are pushed."""
        return self.key() in ARROW_KEYS

    def is_moving_func(self) -> bool:
        """True if function arrows are pushed."""
        return self.key() in FUNC_ARROW_KEYS

    def has_modifier(self, modifier: constants.KeyboardModifierStr) -> bool:
        """True if keycombo contains modifier."""
        return bool(self.keyboardModifiers() & constants.KEYBOARD_MODIFIERS[modifier])

    def has_key(self) -> bool:
        """True if non-modifier key is pressed."""
        return self.key() != Keys.No

    def get_key(self) -> constants.KeyStr:
        return constants.KEY.inverse[self.key()]

    def get_modifiers(self) -> list[constants.KeyboardModifierStr]:
        return constants.KEYBOARD_MODIFIERS.get_list(self.keyboardModifiers())


if __name__ == "__main__":
    seq = KeyCombination(Key.Key_A)
    seq + seq
    print(seq + seq)



from __future__ import annotations

from typing import overload

from prettyqt import animations, widgets
from prettyqt.utils import listdelegators


class StackedWidget(widgets.FrameMixin, widgets.QStackedWidget):
    """Widget containing stack of widgets where only one widget is visible at a time."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.animator = animations.Animator(self)

    def __add__(self, other: widgets.QWidget) -> StackedWidget:
        self.addWidget(other)
        return self

    @overload
    def __getitem__(self, index: int) -> widgets.QWidget: ...

    @overload
    def __getitem__(
        self, index: slice
    ) -> listdelegators.ListDelegator[widgets.QWidget]: ...

    def __getitem__(
        self, index: int | slice
    ) -> widgets.QWidget | listdelegators.ListDelegator[widgets.QWidget]:
        match index:
            case int():
                if index >= self.count():
                    raise IndexError(index)
                w = self.widget(index)
                assert w is not None
                return w
            case slice():
                rng = range(index.start or 0, index.stop or self.count(), index.step or 1)
                return listdelegators.ListDelegator(self.widget(i) for i in rng)
            case _:
                raise TypeError(index)

    def __delitem__(self, item: int | widgets.QWidget):
        i = self.widget(item) if isinstance(item, int) else item
        self.removeWidget(i)

    def __len__(self):
        # needed for PySide6
        return self.count()

    def __contains__(self, item: widgets.QWidget):
        return self.indexOf(item) >= 0

    def set_current_widget(self, widget: widgets.QWidget | int):
        if isinstance(widget, widgets.QWidget):
            self.setCurrentWidget(widget)
        else:
            w = self[widget]
            self.setCurrentWidget(w)

    # __iter__ not needed, we have __getitem__ and __len__


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    stackedwidget = StackedWidget()
    widget2 = widgets.RadioButton("Test")
    widget3 = widgets.PlainTextEdit("Test 243434")
    stackedwidget += widget2
    stackedwidget += widget3
    for i in stackedwidget:
        print(i)
    stackedwidget.show()
    app.exec()



from __future__ import annotations

import logging

from prettyqt import constants, core, custom_widgets, gui, itemmodels, widgets
from prettyqt.utils import signallogger


logger = logging.getLogger(__name__)


def color_log(x):
    match x:
        case "DEBUG":
            return gui.QColor("lightblue")
        case "INFO":
            return gui.QColor("lightgreen")
        case "WARNING":
            return gui.QColor("orange")
        case "ERROR":
            return gui.QColor("red")


class LogRecordTableView(widgets.TableView):
    """Table view showing a log table."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_selection_behavior("rows")
        self.handler = signallogger.SignalLogger()
        core.CoreApplication.call_on_exit(lambda: logger.removeHandler(self.handler))
        self.setWordWrap(True)

    def set_logger(self, logger, level=logging.DEBUG):
        model = itemmodels.LogRecordModel(parent=self)
        logger.addHandler(self.handler)
        self.handler.signals.log_record.connect(model.add)
        self.handler.setLevel(level)
        self.set_model(model)
        self.proxifier[0].modify(color_log, role=constants.BACKGROUND_ROLE)
        self.h_header = custom_widgets.FilterHeader(self)


if __name__ == "__main__":
    from prettyqt import debugging

    app = widgets.app()
    widget = widgets.LineEdit()
    widget.show()
    with app.debug_mode(), debugging.Stalker(widget, log_level=logging.INFO) as stalker:
        stalker.eventsignals.MouseButtonPress.connect(print)
        stalker.show()
        app.exec()



"""Widgets for displaying SVG files."""

from prettyqt.qt.QtSvgWidgets import *  # noqa: F403

from .graphicssvgitem import GraphicsSvgItem
from .svgwidget import SvgWidget
from prettyqt.qt import QtSvgWidgets

QT_MODULE = QtSvgWidgets

__all__ = ["GraphicsSvgItem", "SvgWidget"]



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtDesigner


class DesignerFormEditorInterface(
    core.ObjectMixin, QtDesigner.QDesignerFormEditorInterface
):
    """Allows you to access Qt Designer's various components."""


if __name__ == "__main__":
    interface = DesignerFormEditorInterface()



from __future__ import annotations

from typing import Literal

from prettyqt import widgets
from prettyqt.utils import bidict


mod = widgets.QStyleOptionToolBox


SelectedPositionStr = Literal["not_adjacent", "next_is_selected", "previous_is_selected"]

SELECTED_POSITION: bidict[SelectedPositionStr, mod.SelectedPosition] = bidict(
    not_adjacent=mod.SelectedPosition.NotAdjacent,
    next_is_selected=mod.SelectedPosition.NextIsSelected,
    previous_is_selected=mod.SelectedPosition.PreviousIsSelected,
)


TabPositionStr = Literal["beginning", "middle", "end", "only_one_tab"]

TAB_POSITION: bidict[TabPositionStr, mod.TabPosition] = bidict(
    beginning=mod.TabPosition.Beginning,
    middle=mod.TabPosition.Middle,
    end=mod.TabPosition.End,
    only_one_tab=mod.TabPosition.OnlyOneTab,
)


class StyleOptionToolBox(widgets.StyleOptionMixin, widgets.QStyleOptionToolBox):
    def set_selected_position(self, position: SelectedPositionStr):
        self.selectedPosition = SELECTED_POSITION[position]

    def get_selected_position(self) -> SelectedPositionStr:
        return SELECTED_POSITION.inverse[self.selectedPosition]


if __name__ == "__main__":
    opt = StyleOptionToolBox()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, ClassVar

import attr
import attrs

from prettyqt import constants, gui, itemmodels
from prettyqt.utils import datatypes


if TYPE_CHECKING:
    from prettyqt import core


logger = logging.getLogger(__name__)


class AttrsFieldsModel(itemmodels.BaseFieldsModel):
    """Table model to display the fields and their metadata of an dataclass.

    More information about attrs can be found [here](https://www.attrs.org/).

    Frozen dataclasses are read-only, otherwise the data can be modified.
    """

    SUPPORTS = datatypes.IsAttrs
    HEADER: ClassVar = [
        "Value",
        "Type",
        "Default",
        "Metadata",
        "In __init__",
        "In __repr__",
        "Eq",
        "Hash",
        "Keyword only",
        "Inherited",
        "Validator",
    ]

    def __init__(self, instance: datatypes.IsAttrs, **kwargs):
        super().__init__(instance, **kwargs)

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, datatypes.IsAttrs)

    def get_fields(self, instance: datatypes.IsAttrs):
        fields = attrs.fields(type(instance))
        return {field.name: field for field in fields}

    def data(  # noqa: PLR0911
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        field_name = self._field_names[index.row()]
        field = self._fields[field_name]
        value = getattr(self._instance, field_name)
        match role, index.column():
            case constants.FONT_ROLE, 0:
                font = gui.QFont()
                font.setBold(True)
                return font
            case constants.DISPLAY_ROLE, 0:
                return repr(value)
            case constants.EDIT_ROLE, 0:
                return value
            case constants.DISPLAY_ROLE, 1:
                return field.type
            case constants.FONT_ROLE, 1:
                font = gui.QFont()
                font.setItalic(True)
                return font
            case constants.DISPLAY_ROLE, 2:
                return field.default
            case constants.DISPLAY_ROLE, 3:
                return str(field.metadata)
            case constants.EDIT_ROLE, 3:
                return field.metadata
            case constants.CHECKSTATE_ROLE, 4:
                return self.to_checkstate(field.init)
            case constants.CHECKSTATE_ROLE, 5:
                return self.to_checkstate(field.repr)
            case constants.CHECKSTATE_ROLE, 6:
                return self.to_checkstate(field.eq)
            case constants.CHECKSTATE_ROLE, 7:
                return self.to_checkstate(field.hash)
            case constants.CHECKSTATE_ROLE, 8:
                return self.to_checkstate(field.kw_only)
            case constants.CHECKSTATE_ROLE, 9:
                return self.to_checkstate(field.inherited)
            case constants.CHECKSTATE_ROLE, 10:
                return self.to_checkstate(field.validator is not None)
            case constants.USER_ROLE, _:
                return value

    def _is_writable(self, field_name: str) -> bool:
        # return all(
        #     base_cls.__setattr__ is not attr._make._frozen_setattrs
        #     for base_cls in type(self._instance).__bases__
        # )
        return not attr._make._has_frozen_base_class(type(self._instance))

    def flags(self, index: core.ModelIndex):
        match index.column():
            case 0 | 1 | 2 | 3:
                return super().flags(index)
            case _:
                return super().flags(index) & ~constants.IS_ENABLED


if __name__ == "__main__":
    from prettyqt import widgets

    @attrs.define
    class SelectionSetting:
        name: str
        label: str
        description: str
        requires_restart: bool = False
        options: list[str] | None = None
        minimum: float = 0.0
        maximum: float = 0.0
        pattern: str = ""

    app_style = SelectionSetting(
        name="app_style",
        label="App style",
        description="Some longer text",
        options=[],
        requires_restart=True,
    )
    app = widgets.app()
    view = widgets.TableView()
    view.set_icon("mdi.folder")
    with app.debug_mode():
        model = AttrsFieldsModel(app_style)
        print(dir(app_style))
        print(dir(app_style.__attrs_attrs__))
        model.dataChanged.connect(view.repaint)
        view.set_model(model)
        view.set_selection_behavior("rows")
        view.setEditTriggers(view.EditTrigger.AllEditTriggers)
        view.set_delegate("editor", column=0)
        view.show()
        view.resize(500, 300)
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core, widgets


AnchorStr = Literal["center", "top", "left", "right", "bottom"]


class ZoomAnimation(core.ParallelAnimationGroup):
    ID = "zoom"

    def __init__(
        self,
        duration: int = 1000,
        start: float = 1.0,
        end: float = 1.0,
        easing: core.easingcurve.TypeStr = "in_out_sine",
        anchor: AnchorStr = "center",
        parent: core.QObject | None = None,
    ):
        self._start = start
        self._end = end
        self._anchor = anchor
        super().__init__(parent)
        # size change animation
        self.anim1 = core.PropertyAnimation()
        # slide animation
        self.anim2 = core.PropertyAnimation()
        self.apply_to(parent)
        self.set_easing(easing)
        self.set_start_value(start)
        self.set_end_value(end)
        self.set_duration(duration)
        self.addAnimation(self.anim1)
        self.addAnimation(self.anim2)

    def set_start_value(self, factor: float):
        size = self.parent().size()
        self._start = factor
        self.anim1.setStartValue(size)
        self.anim2.setStartValue(self.parent().pos())

    def set_end_value(self, factor: float):
        self._end = factor
        zoom_ratio = self._end / self._start
        # setup zoom
        size = self.parent().size()
        new = size.toSizeF() * zoom_ratio * zoom_ratio
        self.anim1.setEndValue(new.toSize())
        # setup move
        move_x = -zoom_ratio * self.parent().width()
        move_y = -zoom_ratio * self.parent().height()
        match self._anchor:
            case "center":
                movement = core.Point(int(move_x / 2), int(move_y / 2))
            case "left":
                movement = core.Point(0, int(move_y / 2))
            case "right":
                movement = core.Point(int(move_x), int(move_y / 2))
            case "top":
                movement = core.Point(int(move_x / 2), 0)
            case "bottom":
                movement = core.Point(int(move_x / 2), int(move_y))
        self.anim2.setEndValue(self.parent().pos() + movement)

    def set_duration(self, duration: int):
        self.anim1.setDuration(duration)
        self.anim2.setDuration(duration)

    def set_easing(self, easing: core.easingcurve.TypeStr | core.QEasingCurve.Type):
        self.anim1.set_easing(easing)
        self.anim2.set_easing(easing)

    def apply_to(self, obj: widgets.QWidget):
        self.anim1.apply_to(obj.size)
        self.anim2.apply_to(obj.pos)

    def start(self, *args, **kwargs):
        # update values before starting
        self.set_start_value(self._start)
        self.set_end_value(self._end)
        super().start(*args, **kwargs)


if __name__ == "__main__":
    app = widgets.app()
    val = ZoomAnimation()
    w = widgets.RadioButton()
    val.apply_to(w)
    val.start()
    w.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, widgets
from prettyqt.utils import get_repr


class GraphicsEllipseItem(
    widgets.AbstractGraphicsShapeItemMixin, widgets.QGraphicsEllipseItem
):
    """Ellipse item that you can add to a QGraphicsScene."""

    def __repr__(self):
        return get_repr(self, self.get_rect())

    def get_rect(self) -> core.RectF:
        return core.RectF(self.rect())


if __name__ == "__main__":
    item = GraphicsEllipseItem()
    item.get_rect()



from __future__ import annotations

from prettyqt import gui


class ColorValidator(gui.Validator):
    """Validator which checks for strings representing a color."""

    ID = "color"

    def __eq__(self, other: object):
        return isinstance(other, ColorValidator)

    def validate(self, text: str, pos: int = 0) -> tuple[gui.QValidator.State, str, int]:
        color = gui.Color(text)
        if color.isValid():
            return self.State.Acceptable, text, pos
        return self.State.Intermediate, text, pos


if __name__ == "__main__":
    from prettyqt import widgets

    val = ColorValidator()
    app = widgets.app()
    widget = widgets.LineEdit()
    widget.setValidator(val)
    widget.show()
    app.exec()



"""Classes for integrating documentation into applications."""

from __future__ import annotations

from prettyqt.qt.QtHelp import *  # noqa: F403

from .helplink import HelpLink
from .helpindexmodel import HelpIndexModel
from .helpfilterdata import HelpFilterData
from .helpcontentitem import HelpContentItem
from .helpsearchresult import HelpSearchResult
from .helpcontentmodel import HelpContentModel
from .helpcontentwidget import HelpContentWidget
from .helpindexwidget import HelpIndexWidget
from .helpsearchresultwidget import HelpSearchResultWidget
from .helpsearchquerywidget import HelpSearchQueryWidget
from .helpfiltersettingswidget import HelpFilterSettingsWidget
from .helpenginecore import HelpEngineCore, HelpEngineCoreMixin
from .helpengine import HelpEngine
from .helpsearchengine import HelpSearchEngine
from .helpfilterengine import HelpFilterEngine
from prettyqt.qt import QtHelp

QT_MODULE = QtHelp

__all__ = [
    "HelpContentItem",
    "HelpContentModel",
    "HelpContentWidget",
    "HelpEngine",
    "HelpEngineCore",
    "HelpEngineCoreMixin",
    "HelpFilterData",
    "HelpFilterEngine",
    "HelpFilterSettingsWidget",
    "HelpIndexModel",
    "HelpIndexWidget",
    "HelpLink",
    "HelpSearchEngine",
    "HelpSearchQueryWidget",
    "HelpSearchResult",
    "HelpSearchResultWidget",
]



from __future__ import annotations

import collections
import contextlib

# import enum
import logging
from typing import Any, TypeVar

from prettyqt import constants, core, eventfilters, gui


logger = logging.getLogger(__name__)

T = TypeVar("T", bound=core.QObject)


class EventSignaller(core.Object):
    signal = core.Signal(core.QEvent)

    def __init__(self, obj: core.QObject, **kwargs):
        super().__init__(**kwargs)
        self._obj = obj

    def __getattr__(self, name):
        typ = getattr(core.QEvent.Type, name)
        self.ec = self._obj.add_callback_for_event(self._on_event_detected, include=typ)
        return self.signal

    def _on_event_detected(self, event: core.QEvent):
        self.signal.emit(event)
        return False


class Stalker(core.Object):
    # @core.Enum
    # class LogLevel(enum.IntEnum):
    #     """Log level."""

    #     DEBUG = logging.DEBUG
    #     INFO = logging.INFO
    #     WARNING = logging.WARNING
    #     CRITICAL = logging.CRITICAL
    #     ERROR = logging.ERROR

    keypress_detected = core.Signal(str)
    leftclick_detected = core.Signal(core.QPointF)
    rightclick_detected = core.Signal(core.QPointF)
    event_detected = core.Signal(core.QEvent)
    signal_emitted = core.Signal(core.MetaMethod, object)  # signal, args
    signal_connected = core.Signal(core.MetaMethod)
    signal_disconnected = core.Signal(core.MetaMethod)

    def __init__(
        self,
        qobject: core.QObject,
        include=None,
        exclude=None,
        **kwargs: Any,
    ):
        self._log_level = logging.INFO
        super().__init__(**kwargs)
        self._obj = qobject
        qmeta = self._obj.metaObject()
        assert qmeta
        self._meta = core.MetaObject(qmeta)
        self.counter: collections.defaultdict[str, int] = collections.defaultdict(int)
        self.signal_counter: collections.defaultdict[str, int] = collections.defaultdict(
            int
        )
        self.exclude = ["meta_call", "timer"] if exclude is None else exclude
        self.include = include
        self._handles: list[core.QMetaObject.Connection] = []

    def __enter__(self):
        self.hook()
        return self

    def __exit__(self, typ, value, traceback):
        self.unhook()

    @property
    def eventsignals(self):
        return EventSignaller(self._obj)

    def hook(self):
        # enable event logging by installing EventCatcher, which includes logging
        self.eventcatcher = eventfilters.EventCatcher(
            self.include, self.exclude, self._on_event_detected, parent=self._obj
        )
        self._obj.installEventFilter(self.eventcatcher)
        # enable logging of signals emitted by connecting all signals to our fn
        for signal in self._meta.get_signals(only_notifiers=False):
            signal_name = signal.get_name()
            # PyQt reports non-existing signals in MetaObject.
            if hasattr(self._obj, signal_name):
                signal_instance = getattr(self._obj, signal_name)
            fn = self._on_signal_emitted(signal)
            handle = signal_instance.connect(fn)
            self._handles.append(handle)
        self.log(f"Stalking {len(self._handles)} signals")
        # enable logging of all signal (dis)connections by hooking to connectNotify
        self.old_connectNotify = self._obj.connectNotify
        self.old_disconnectNotify = self._obj.disconnectNotify
        self._obj.connectNotify = self._on_signal_connected
        self._obj.disconnectNotify = self._on_signal_disconnected
        # self._obj.destroyed.connect(self.unhook)
        # self.destroyed.connect(self.unhook)

    def unhook(self):
        if self.eventcatcher is None:
            logger.warning("unhook() called before hook()")
            return
        """Clean up our mess."""
        self._obj.connectNotify = self.old_connectNotify
        self._obj.disconnectNotify = self.old_disconnectNotify
        self.old_connectNotify = None
        self.old_disconnectNotify = None
        for handle in self._handles:
            self._obj.disconnect(handle)
        self._handles = []
        with contextlib.suppress(RuntimeError):
            self._obj.removeEventFilter(self.eventcatcher)

    def log(self, message: str):
        if self.log_level:
            with contextlib.suppress(RuntimeError):
                logger.log(self._log_level, f"{self._obj!r}: {message}")  # noqa: G004

    def _on_signal_connected(self, qsignal: core.QMetaMethod):
        signal = core.MetaMethod(qsignal)
        self.log(f"Connected signal {signal.get_name()}")
        self.signal_connected.emit(signal)

    def _on_signal_disconnected(self, qsignal: core.QMetaMethod):
        signal = core.MetaMethod(qsignal)
        self.log(f"Disconnected signal {signal.get_name()}")
        self.signal_disconnected.emit(signal)

    def _on_event_detected(self, event) -> bool:
        """Used for EventCatcher, returns false to not eat signals."""
        try:
            self.event_detected.emit(event)
        except RuntimeError:
            return None
        match event.type():
            case core.Event.Type.KeyPress:
                combo = gui.KeySequence(event.keyCombination()).toString()
                self.keypress_detected.emit(combo)
            case core.Event.Type.MouseButtonRelease:
                if event.button() == constants.MouseButton.LeftButton:
                    self.leftclick_detected.emit(event.position())
                if event.button() == constants.MouseButton.RightButton:
                    self.rightclick_detected.emit(event.position())
        self.log(f"Received event {event.type()!r}")
        self.counter[event.type()] += 1
        return False

    def _on_signal_emitted(self, signal: core.MetaMethod):
        def fn(*args, **kwargs):
            try:
                self.signal_emitted.emit(signal, args)
                self.signal_counter[signal.get_name()] += 1
                self.log(f"Emitted signal {signal.get_name()}{args}")
            except RuntimeError:
                pass

        return fn

    def set_log_level(self, level: int):
        self._log_level = level

    def get_log_level(self) -> int:
        return self._log_level

    def count_children(self, type_filter: type[T] = core.QObject) -> collections.Counter:
        objects = self.findChildren(type_filter)
        return collections.Counter([type(o) for o in objects])

    def show(self):
        from prettyqt import custom_widgets

        widget = custom_widgets.LogRecordTableView()
        widget.set_logger(logger, level=self.log_level)
        widget.show()

    log_level = core.Property(
        int,
        get_log_level,
        set_log_level,
        doc="Level to use for logging",
    )


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = widgets.LineEdit()
    widget.show()
    with app.debug_mode(), Stalker(widget, log_level=logging.INFO) as stalker:
        stalker.eventsignals.MouseButtonPress.connect(print)
        stalker.show()
        app.exec()



from __future__ import annotations

import logging
import os
import pathlib
import sys
from typing import TYPE_CHECKING, ClassVar

from prettyqt import constants, core


if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine

    from prettyqt.utils import datatypes


logger = logging.getLogger(__name__)


class CoreApplicationMixin(core.ObjectMixin):
    translators: ClassVar[dict[str, core.Translator]] = {}

    @classmethod
    def call_on_exit(cls, func: Callable):
        instance = cls.instance()
        if instance is None:
            msg = "No QApplication running"
            raise RuntimeError(msg)
        instance.aboutToQuit.connect(func)

    @classmethod
    def get_application_file_path(cls) -> pathlib.Path:
        return pathlib.Path(cls.applicationFilePath())

    @classmethod
    def get_application_dir_path(cls) -> pathlib.Path:
        return pathlib.Path(cls.applicationDirPath())

    @classmethod
    def add_library_path(cls, path: datatypes.PathType):
        cls.addLibraryPath(os.fspath(path))

    @classmethod
    def get_library_paths(cls) -> list[pathlib.Path]:
        return [pathlib.Path(i) for i in cls.libraryPaths()]

    def set_application_name(self, name: str):
        if os.name == "nt" and name and not getattr(sys, "frozen", False):
            import ctypes

            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(name)
        self.setApplicationName(name)

    def set_metadata(
        self,
        app_name: str | None = None,
        app_version: None | datatypes.SemanticVersionType = None,
        org_name: str | None = None,
        org_domain: str | None = None,
    ):
        if app_name is not None:
            self.setApplicationName(app_name)
        match app_version:
            case None:
                pass
            case core.QVersionNumber():
                app_version = app_version.toString()
                self.setApplicationVersion(app_version)
            case tuple():
                app_version = ".".join(str(i) for i in app_version)
                self.setApplicationVersion(app_version)
            case str():
                self.setApplicationVersion(app_version)
            case _:
                raise TypeError(app_version)
        if org_name is not None:
            self.setOrganizationName(org_name)
        if org_domain is not None:
            self.setOrganizationDomain(org_domain)

    @classmethod
    def load_language_file(cls, file: datatypes.PathType) -> core.Translator:
        translator = core.Translator()
        translator.load_file(file)
        cls.installTranslator(translator)
        cls.translators[str(file)] = translator
        return translator

    @classmethod
    def load_language(cls, language: str) -> core.Translator:
        translator = core.Translator.for_language(language)
        cls.installTranslator(translator)
        cls.translators[language] = translator
        return translator

    def post_event(
        self,
        obj: core.QObject,
        event: core.QEvent,
        priority: int | constants.EventPriorityStr = "normal",
    ):
        match priority:
            case int():
                prio = priority
            case str():
                prio = constants.EVENT_PRIORITY[priority]
            case _:
                raise TypeError(priority)
        return self.postEvent(obj, event, prio)

    def in_main_thread(self) -> bool:
        """Check if we are in the thread in which QApplication object was created.

        Returns:
            True if we are in the main thread, False otherwise.
        """
        return self.thread() == core.Thread.currentThread()

    def main_loop(self) -> int:
        return self.exec()

    @staticmethod
    def run_async(
        coro: Coroutine | None = None,
        keep_running: bool = True,
        quit_qapp: bool = True,
        *,
        handle_sigint: bool = False,
        debug: bool | None = None,
    ) -> int:
        from prettyqt.qt import QtAsyncio

        return QtAsyncio.run(
            coro=coro,
            keep_running=keep_running,
            quit_qapp=quit_qapp,
            handle_sigint=handle_sigint,
            debug=debug,
        )

    @staticmethod
    def restart():
        os.execl(sys.executable, sys.executable, *sys.argv)
        # process = QProcess()
        # process.setProgram(sys.executable)

        # if not running_as_bundled_app():
        #     process.setArguments(sys.argv)

        # process.startDetached()
        # self.close(quit_app=True)


class CoreApplication(CoreApplicationMixin, core.QCoreApplication):
    """Event loop for Qt applications without UI."""


if __name__ == "__main__":
    app = CoreApplication([])
    app.applicationNameChanged.connect(print)



from __future__ import annotations

from prettyqt import constants, widgets


class GraphicsLinearLayout(widgets.GraphicsLayoutMixin, widgets.QGraphicsLinearLayout):
    """Horizontal or vertical layout for managing widgets in Graphics View."""

    def __init__(
        self,
        orientation: constants.OrientationStr | constants.Orientation = "horizontal",
        parent: widgets.QGraphicsLayoutItem | None = None,
    ):
        ori = constants.ORIENTATION.get_enum_value(orientation)
        super().__init__(ori, parent)

    def __add__(self, other):
        self[self.count()] = other
        return self


if __name__ == "__main__":
    app = widgets.app()
    layout = GraphicsLinearLayout()



from __future__ import annotations

from prettyqt import constants, gui, widgets


class GraphicsPolygonItem(
    widgets.AbstractGraphicsShapeItemMixin, widgets.QGraphicsPolygonItem
):
    """Polygon item that you can add to a QGraphicsScene."""

    def get_polygon(self) -> gui.PolygonF:
        return gui.PolygonF(self.polygon())

    def set_fill_rule(self, rule: constants.FillRuleStr | constants.FillRule):
        self.setFillRule(constants.FILL_RULE.get_enum_value(rule))

    def get_fill_rule(self) -> constants.FillRuleStr:
        return constants.FILL_RULE.inverse[self.fillRule()]



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from prettyqt import widgets
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


mod = widgets.QInputDialog

InputDialogOptionStr = Literal[
    "no_buttons", "use_listview_for_combobox_items", "use_plaintextedit_for_text_input"
]

INPUT_DIALOG_OPTION: bidict[InputDialogOptionStr, mod.InputDialogOption] = bidict(
    no_buttons=mod.InputDialogOption.NoButtons,
    use_listview_for_combobox_items=mod.InputDialogOption.UseListViewForComboBoxItems,
    use_plaintextedit_for_text_input=mod.InputDialogOption.UsePlainTextEditForTextInput,
)

InputModeStr = Literal["text", "int", "double"]

INPUT_MODE: bidict[InputModeStr, mod.InputMode] = bidict(
    text=mod.InputMode.TextInput,
    int=mod.InputMode.IntInput,
    double=mod.InputMode.DoubleInput,
)


class InputDialog(widgets.DialogMixin, widgets.QInputDialog):
    """Simple convenience dialog to get a single value from the user."""

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "textEchoMode": widgets.lineedit.ECHO_MODE,
            "inputMode": INPUT_MODE,
            "options": INPUT_DIALOG_OPTION,
        }
        return maps

    @classmethod
    def get_int(
        cls,
        title: str = "",
        label: str = "",
        icon: datatypes.IconType = None,
    ) -> int | None:
        par = widgets.Dialog()
        par.set_icon(icon)
        v = cls.getInt(par, title, label, value=0)
        return v[0] if v[1] else None

    @classmethod
    def get_float(
        cls,
        title: str = "",
        label: str = "",
        icon: datatypes.IconType = None,
    ) -> float | None:
        par = widgets.Dialog()
        par.set_icon(icon)
        v = cls.getDouble(par, title, label, value=0.0)
        return v[0] if v[1] else None

    @classmethod
    def get_text(
        cls,
        title: str = "",
        label: str = "",
        icon: datatypes.IconType = None,
        value: str = "",
        echo_mode: widgets.lineedit.EchoModeStr | widgets.QLineEdit.EchoMode = "normal",
    ) -> str | None:
        par = widgets.Dialog()
        par.set_icon(icon)
        v = cls.getText(
            par, title, label, widgets.lineedit.ECHO_MODE.get_enum_value(echo_mode), value
        )
        return v[0] if v[1] else None

    @classmethod
    def get_item(
        cls,
        items: list[str],
        title: str = "",
        label: str = "",
        icon: datatypes.IconType = None,
        editable: bool = False,
    ) -> str | None:
        par = widgets.Dialog()
        par.set_icon(icon)
        v = cls.getItem(par, title, label, items, editable=editable, current=0)
        return v[0] if v[1] else None

    def set_text_echo_mode(
        self, mode: widgets.lineedit.EchoModeStr | widgets.QLineEdit.EchoMode
    ):
        """Set text echo mode.

        Args:
            mode: echo mode to use
        """
        self.setTextEchoMode(widgets.lineedit.ECHO_MODE.get_enum_value(mode))

    def get_text_echo_mode(self) -> widgets.lineedit.EchoModeStr:
        """Return text echo mode.

        Returns:
            echo mode
        """
        return widgets.lineedit.ECHO_MODE.inverse[self.textEchoMode()]

    def set_input_mode(self, mode: InputModeStr | mod.InputMode):
        """Set input mode.

        Args:
            mode: input mode to use
        """
        self.setInputMode(INPUT_MODE.get_enum_value(mode))

    def get_input_mode(self) -> InputModeStr:
        """Return input mode.

        Returns:
            input mode
        """
        return INPUT_MODE.inverse[self.inputMode()]


if __name__ == "__main__":
    app = widgets.app()
    result = InputDialog.get_int("a", "b", icon="mdi.timer")
    app.exec()



from __future__ import annotations

from prettyqt import core


class PluginLoader(core.ObjectMixin, core.QPluginLoader):
    """Load a plugin at runtime."""

    def get_load_hints(self) -> list[core.library.LoadHintStr]:
        return core.library.LOAD_HINTS.get_list(self.loadHints())

    def set_load_hints(self, **kwargs):
        flag = core.QLibrary.LoadHint(0)  # type: ignore
        for k, v in kwargs.items():
            if v is True:
                flag |= core.library.LOAD_HINTS[k]  # type: ignore
        self.setLoadHints(flag)  # type: ignore



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from prettyqt import core


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class FileSelector(core.ObjectMixin, core.QFileSelector):
    """Convenient way of selecting file variants."""

    def select_path(self, path: datatypes.PathType) -> str:
        return self.select(os.fspath(path))

    def select_url(self, url: datatypes.UrlType) -> core.Url:
        if isinstance(url, str):
            url = core.QUrl(url)
        return core.Url(self.select(url))



from __future__ import annotations

import contextlib
import logging
import os
import pathlib
import sys
import timeit
from typing import TYPE_CHECKING

import qstylizer.parser
import qstylizer.style

from prettyqt import constants, core, gui, iconprovider, paths, widgets
from prettyqt.utils import datatypes


if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator, Mapping, MutableMapping

    from prettyqt.utils import listdelegators


logger = logging.getLogger(__name__)


SAVE_STATES = dict(
    splitters=widgets.QSplitter,
    mainwindows=widgets.QMainWindow,
    headerviews=widgets.QHeaderView,
)


def setup_runner():
    import inspect

    from prettyqt.qt.QtCore import SignalInstance
    from prettyqt.utils import asyncrunner

    old_connect = SignalInstance.connect
    runner = asyncrunner.AsyncRunner()

    def connect(self, slot):
        if inspect.iscoroutinefunction(slot):
            return old_connect(self, runner.to_sync(slot))
        return old_connect(self, slot)

    SignalInstance.connect = connect
    return runner


class ApplicationMixin(gui.GuiApplicationMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._debug = False
        self.runner = setup_runner()

    def __iter__(self) -> Iterator[widgets.QWidget]:
        return iter(self.topLevelWidgets())

    def run_parallel(self, fns: Iterable[Callable]):
        self.runner.run_parallel(fns)

    def is_debug(self) -> bool:
        return self._debug

    @contextlib.contextmanager
    def debug_mode(self, log_level: int = logging.DEBUG):
        from prettyqt.debugging import ErrorMessageBox, MessageHandler
        from prettyqt.eventfilters import debugmode

        handler = logging.StreamHandler(sys.stdout)
        f_format = logging.Formatter(
            "%(asctime)s: %(filename)s:%(lineno)d - %(levelname)s - %(message)s"
        )
        handler.setFormatter(f_format)
        handler.setLevel(log_level)
        root_logger = logging.getLogger()
        root_logger.addHandler(handler)
        root_logger.setLevel(log_level)
        _original_excepthook = sys.excepthook
        sys.excepthook = ErrorMessageBox._excepthook
        eventfilter = debugmode.DebugMode(self)
        self._debug = True
        self.installEventFilter(eventfilter)
        with MessageHandler(root_logger):
            yield self
        self._debug = False
        self.removeEventFilter(eventfilter)
        sys.excepthook = _original_excepthook

    def store_widget_states(
        self, settings: MutableMapping | None = None, key: str = "states"
    ):
        settings = core.Settings() if settings is None else settings
        result = {
            k: {
                i.objectName(): i.saveState()
                for i in self.find_children(v)
                if i.objectName()
            }
            for k, v in SAVE_STATES.items()
        }
        settings[key] = result

    def restore_widget_states(self, settings: Mapping | None = None, key: str = "states"):
        settings = core.Settings() if settings is None else settings
        for category, v in SAVE_STATES.items():
            items = settings[key].get(category)
            if items is None:
                continue
            for name, state in items.items():
                w = self.find_child(v, name=name)
                if w is not None:
                    new_state = state.encode() if isinstance(state, str) else state
                    w.restoreState(new_state)

    @classmethod
    def widget_at(
        cls, pos: datatypes.PointType, typ: type | None = None
    ) -> widgets.QWidget:
        if typ is None:
            return super().widgetAt(pos)
        for widget in cls.widgets_at(pos):
            if isinstance(widget, typ):
                return widget
        return None

    @classmethod
    def widgets_at(
        cls, pos: datatypes.PointType
    ) -> listdelegators.ListDelegator[widgets.QWidget]:
        """Return ALL widgets at `pos`.

        Arguments:
            pos: Position at which to get widgets

        Returns:
            list of widgets at given position

        """
        widgets = []
        while widget_at := cls.widgetAt(datatypes.to_point(pos)):
            widgets.append(widget_at)
            # Make widget invisible to further enquiries
            widget_at.setAttribute(constants.WidgetAttribute.WA_TransparentForMouseEvents)
        # Restore attribute
        for widget in widgets:
            widget.setAttribute(
                constants.WidgetAttribute.WA_TransparentForMouseEvents, False
            )

        return widgets

    @classmethod
    def get_mainwindow(cls) -> widgets.QMainWindow | None:
        app = cls.instance()
        if app is None:
            return None
        return next(
            (w for w in app.topLevelWidgets() if isinstance(w, widgets.QMainWindow)),
            None,
        )

    @classmethod
    def get_widget(cls, name: str) -> widgets.QWidget | None:
        mw = cls.get_mainwindow()
        if mw is None:
            logger.warning("Trying to get widget from nonexistent mainwindow")
            return None
        return mw.findChild(widgets.QWidget, name)  # type: ignore
        # widget_list = cls.instance().allWidgets()
        # for widget in widget_list:
        #     if isinstance(widget, widgets.QWidget) and widget.objectName() == name:
        #         return widget
        # return None

    @contextlib.contextmanager
    def edit_stylesheet(self) -> Iterator[qstylizer.style.StyleSheet]:
        ss = self.get_stylesheet()
        yield ss
        self.set_stylesheet(ss)

    def set_stylesheet(
        self, ss: None | str | qstylizer.style.StyleSheet | datatypes.PathType
    ):
        match ss:
            case str():
                pass
            case os.PathLike():
                ss = pathlib.Path(ss).read_text("utf-8")
            case None:
                ss = ""
            case qstylizer.style.StyleSheet():
                ss = str(ss)
            case _:
                raise TypeError(ss)
        self.setStyleSheet(ss)

    def set_style(self, style: str):
        self.setStyle(widgets.QStyleFactory.create(style))
        icon_color = self.get_palette().get_color("highlighted_text")
        iconprovider.set_defaults(color=icon_color)

    def get_stylesheet(self) -> qstylizer.style.StyleSheet:
        return qstylizer.parser.parse(self.styleSheet())

    def set_theme(self, theme: constants.ThemeStr):
        self.set_palette(theme)
        match theme:
            case "default":
                self.set_stylesheet("")
                color = self.get_palette().get_color("highlighted_text")
            case "dark":
                ss = (paths.THEMES_PATH / "darktheme.qss").read_text("utf-8")
                self.set_stylesheet(ss)
                color = gui.Color("lightblue")
        iconprovider.set_defaults(color=color)

    @classmethod
    def get_available_themes(cls) -> dict[constants.ThemeStr, str]:
        return dict(default="Default", dark="Dark")

    def send_event(self, obj_or_str: str | core.QObject, event: core.QEvent) -> bool:
        obj = self.get_widget(obj_or_str) if isinstance(obj_or_str, str) else obj_or_str
        if obj is None:
            raise ValueError(obj)
        return self.sendEvent(obj, event)

    def post_event(
        self,
        obj_or_str: str | core.QObject,
        event: core.QEvent,
        priority: int | constants.EventPriorityStr = "normal",
    ):
        obj = self.get_widget(obj_or_str) if isinstance(obj_or_str, str) else obj_or_str
        if obj is None:
            raise ValueError(obj)
        super().post_event(obj, event, priority)

    @classmethod
    def get_style_icon(
        cls, icon: widgets.style.StandardPixmapStr | widgets.QStyle.StandardPixmap
    ) -> gui.Icon:
        style = cls.style()
        # icon_size = style.pixelMetric(widgets.QStyle.PM_MessageBoxIconSize)
        icon = style.standardIcon(widgets.style.STANDARD_PIXMAP.get_enum_value(icon))
        return gui.Icon(icon)

    def set_effect_enabled(
        self, effect: constants.UiEffectStr | constants.UIEffect, enabled: bool = True
    ):
        """Set the enabled state of a desktop effect.

        Args:
            effect: desktop effect to set
            enabled: new state
        """
        self.setEffectEnabled(constants.UI_EFFECTS.get_enum_value(effect))

    def is_effect_enabled(self, effect: constants.UiEffectStr) -> bool:
        """Return desktop effect state.

        Returns:
            desktop effect state
        """
        return self.isEffectEnabled(constants.UI_EFFECTS[effect])

    def set_navigation_mode(
        self, mode: constants.NavigationModeStr | constants.NavigationMode
    ):
        """Set the navigation mode.

        Args:
            mode: navigation mode to use
        """
        self.setNavigationMode(constants.NAVIGATION_MODES.get_enum_value(mode))

    def get_navigation_mode(self) -> constants.NavigationModeStr:
        """Return navigation mode.

        Returns:
            navigation mode
        """
        return constants.NAVIGATION_MODES.inverse[self.navigationMode()]

    @classmethod
    def sleep(cls, secs: float):
        """Pause application (non-blocking).

        Args:
            secs: seconds to sleep
        """
        start = timeit.default_timer()
        while timeit.default_timer() - start < secs:
            cls.processEvents()

    @classmethod
    def process_events(
        cls,
        user_input: bool = True,
        socket_notifiers: bool = True,
        wait_for_more: bool = True,
        msecs: int = 0,
    ):
        flag = core.EventLoop.ProcessEventsFlag.AllEvents
        if not user_input:
            flag |= core.EventLoop.ProcessEventsFlag.ExcludeUserInputEvents
        if not socket_notifiers:
            flag |= core.EventLoop.ProcessEventsFlag.ExcludeSocketNotifiers
        if wait_for_more:  # doesnt seem to work? Could use this for sleep() otherwise.
            flag |= core.EventLoop.ProcessEventsFlag.WaitForMoreEvents
        cls.processEvents(flag, msecs)


class Application(ApplicationMixin, widgets.QApplication):
    """Manages the GUI application's control flow and main settings."""

    # def __init__(self, *args, **kwargs):
    #     super().__init__()


# Application.setStyle(widgets.Style())

if __name__ == "__main__":
    app = widgets.app()

    container = widgets.Widget()
    container.set_layout("horizontal")
    w = widgets.PlainTextEdit(parent=container)
    w2 = widgets.PlainTextEdit(parent=container)
    container.box.add(w)
    container.box.add(w2)
    container.show()
    # editor = widgeteditor.WidgetEditor(w)
    # editor.show()
    with app.debug_mode():
        app.sleep(1)
        app.exec()
